from steamship_tests import PLUGINS_PATH
from steamship_tests.client.operations.test_embed import (
    basic_embedding_search,
    basic_embeddings,
    count_embeddings,
)
from steamship_tests.client.operations.test_embedding_index import create_index
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship.data import TagKind, TagValue
from steamship.data.plugin import PluginType


def test_e2e_embedder():
    client = get_steamship_client()
    embedder_path = PLUGINS_PATH / "taggers" / "plugin_embedder.py"

    with deploy_plugin(client, embedder_path, PluginType.tagger) as (
        plugin,
        version,
        instance,
    ):
        e1 = instance.tag("This is a test")
        e1.wait()
        assert count_embeddings(e1.output.file) == 1
        assert e1.output.file.blocks[0].tags[0].kind == TagKind.EMBEDDING
        assert len(e1.output.file.blocks[0].tags[0].value[TagValue.VECTOR_VALUE]) > 1

        e2 = instance.tag("This is a test")
        e2.wait()
        assert count_embeddings(e2.output.file) == 1
        assert e2.output.file.blocks[0].tags[0].kind == TagKind.EMBEDDING
        assert len(e2.output.file.blocks[0].tags[0].value[TagValue.VECTOR_VALUE]) == len(
            e1.output.file.blocks[0].tags[0].value[TagValue.VECTOR_VALUE]
        )

        e4 = instance.tag("This is a test")
        e4.wait()
        assert count_embeddings(e4.output.file) == 1

        # Now lets run all the other embedding steamship_tests
        basic_embeddings(instance)
        basic_embedding_search(client, plugin_instance=instance.handle)
        index = create_index(client, plugin_instance=instance.handle)
        index.delete()



import logging
from typing import Type

from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.tagger import Tagger


class TestLoggingTaggerPlugin(Tagger):
    class EmptyConfig(Config):
        pass

    def config_cls(self) -> Type[Config]:
        return self.EmptyConfig

    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        logging.info("A remote logging log")
        file = request.data.file
        if request.data is not None:
            ret = InvocableResponse(data=BlockAndTagPluginOutput(file=file))
            logging.info(f"Ret: {ret}")
            return ret


handler = create_handler(TestLoggingTaggerPlugin)



import pytest
from assets.plugins.taggers.plugin_parser import TestParserPlugin

from steamship import File
from steamship.data.block import Block
from steamship.invocable import InvocableResponse
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

TEST_REQ = BlockAndTagPluginInput(
    file=File(
        blocks=[
            Block(
                id="ABC",
                text="Once upon a time there was a magical ship. "
                "The ship was powered by STEAM. The ship went to the moon.",
            )
        ]
    )
)
TEST_PLUGIN_REQ = PluginRequest(data=TEST_REQ)
TEST_PLUGIN_REQ_DICT = TEST_PLUGIN_REQ.dict()


def _test_resp(res):
    assert isinstance(res, InvocableResponse)
    assert isinstance(res.data, BlockAndTagPluginOutput)
    assert len(res.data.file.blocks) == 1
    assert res.data.file.blocks[0].text == TEST_REQ.file.blocks[0].text
    assert len(res.data.file.blocks[0].tags) == 3


def test_parser():
    parser = TestParserPlugin()
    res = parser.run(TEST_PLUGIN_REQ)
    _test_resp(res)

    # The endpoints take a kwargs block which is transformed into the appropriate JSON object
    res2 = parser.run_endpoint(**TEST_PLUGIN_REQ_DICT)
    _test_resp(res2)

    # This plugin is not trainable, and thus it refuses trainable parameters requests
    with pytest.raises(AttributeError):
        parser.get_training_parameters(PluginRequest(data=TrainingParameterPluginInput()))
    with pytest.raises(AttributeError):
        parser.get_training_parameters_endpoint(
            **PluginRequest(data=TrainingParameterPluginInput()).dict()
        )

    # This plugin is not trainable, and thus it refuses train requests
    with pytest.raises(AttributeError):
        parser.train(PluginRequest(data=TrainPluginInput()))
    with pytest.raises(AttributeError):
        parser.train_endpoint(**PluginRequest(data=TrainPluginInput()).dict())



import logging
import re
from typing import Type

from steamship import Block, DocTag, Tag
from steamship.data import TagKind
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.tagger import Tagger

# If this isn't present, Localstack won't show logs
logging.getLogger().setLevel(logging.INFO)


def tag_sentences(block: Block):
    """Splits the document into sentences by assuming a period is a sentence divider."""
    # Add the period back
    tags = []
    for m in re.finditer(r"[^.]+", block.text):
        tags.append(
            Tag(
                kind=TagKind.DOCUMENT,
                name=DocTag.SENTENCE,
                start_idx=m.start(),
                end_idx=m.end() + 1,
            )
        )
    if block.tags:
        block.tags.extend(tags)
    else:
        block.tags = tags


class TestParserPlugin(Tagger):
    # TODO: WARNING! We will need to implement some logic that prevents
    # a distributed endless loop. E.g., a parser plugin returning the results
    # of using the Steamship client to call parse.. via itself!
    class EmptyConfig(Config):
        pass

    def config_cls(self) -> Type[Config]:
        return self.EmptyConfig

    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        logging.info(f"Inside parser: {type(request)}")
        file = request.data.file
        for block in file.blocks:
            tag_sentences(block)
        if request.data is not None:
            ret = InvocableResponse(data=BlockAndTagPluginOutput(file=file))
            logging.info(f"Ret: {ret}")
            return ret


handler = create_handler(TestParserPlugin)



"""Async CSV Blockifier - Steamship Plugin."""
from typing import Union

from assets.plugins.blockifiers.csv_blockifier import CsvBlockifier

from steamship import SteamshipError
from steamship.base import Task, TaskState
from steamship.invocable import InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.blockifier import Blockifier
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.utils.binary_utils import to_b64

ASYNC_JOB_ID = "bbq-time"
STATUS_MESSAGE = "Still working!"


class AsyncCsvBlockifier(CsvBlockifier, Blockifier):
    """This plugin is identical to the CSV blockifier, but it:
         1) delays its response until the first re-checking of the task status
         2) requires that the status check include the previously returned "async_job_id"

    In so doing, this permits a unit test to verify both the engine's handling of an async blockification
    and also the correct flow of the 'remote_status_input' field.
    """

    def _raw_data_input_to_dict(self, input: RawDataPluginInput) -> dict:
        """Helper function to help us dehydrate the input to the status input."""
        data = to_b64(input.data)
        return {
            "plugin_instance": input.plugin_instance,
            "data": data,
            "default_mime_type": input.default_mime_type,
        }

    def run(
        self, request: PluginRequest[RawDataPluginInput]
    ) -> Union[InvocableResponse, InvocableResponse[BlockAndTagPluginOutput]]:
        if request.is_status_check:
            if request.status.remote_status_input.get("async_job_id") != ASYNC_JOB_ID:
                raise SteamshipError(message="Required status password was not provided")

            # Re-hydrate the input we stashed from before.
            # This wouldn't be necessary in practice -- we're just doing it here in order to construct the task.
            request.data = RawDataPluginInput(**request.status.remote_status_input.get("data"))
            return super().run(request)
        else:
            return InvocableResponse(
                status=Task(
                    state=TaskState.running,
                    remote_status_message=STATUS_MESSAGE,
                    remote_status_input={
                        "async_job_id": ASYNC_JOB_ID,
                        "data": self._raw_data_input_to_dict(request.data),
                    },
                )
            )


handler = create_handler(AsyncCsvBlockifier)



import logging
from pathlib import Path
from typing import Any, Dict, Type

from steamship import File
from steamship.base.client import Client
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.train_status_plugin_input import TrainStatusPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.request import PluginRequest
from steamship.plugin.tagger import TrainableTagger
from steamship.plugin.trainable_model import TrainableModel

# If this isn't present, Localstack won't show logs
logging.getLogger().setLevel(logging.INFO)


class TestConfig(Config):
    test_value1: str = None
    test_value2: str = None


class TestTrainableTaggerConfigModel(TrainableModel[TestConfig]):
    def load_from_folder(self, checkpoint_path: Path):
        assert self.config is not None
        assert self.config.test_value1 is not None
        assert self.config.test_value2 is not None

    def save_to_folder(self, checkpoint_path: Path):
        assert self.config is not None
        assert self.config.test_value1 is not None
        assert self.config.test_value2 is not None

    def train(self, input: TrainPluginInput) -> None:
        assert self.config is not None
        assert self.config.test_value1 is not None
        assert self.config.test_value2 is not None

    def train_status(self, input: TrainStatusPluginInput) -> None:
        assert self.config is not None
        assert self.config.test_value1 is not None
        assert self.config.test_value2 is not None

    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Tags the incoming data for any instance of the keywords in the parameter file."""
        assert self.config is not None
        assert self.config.test_value1 is not None
        assert self.config.test_value2 is not None
        response = InvocableResponse(data=BlockAndTagPluginOutput(file=File.CreateRequest(tags=[])))
        logging.info(f"TestTrainableTaggerModel:run() returning {response}")
        return response


class TestTrainableTaggerConfigPlugin(TrainableTagger):
    """Tests the Trainable Tagger lifecycle.

    - This tagger produces a FIXED set of trainable parameters.
    - These parameters (and not the trainable data!) fully parameterize the trained model.
    - The trained model (and not the trainable parameters!) fully parameterize the running model.
    - The model simply tags keywords that it finds in the text.

    Taken together, this plugin can be seen as a reference implementation of the data/process lifecycle of a trainable
    model, regardless of where the actual work occurs:

    - It could occur here, running in Lambda.
    - It could occur here, running in ECS.
    - It could be orchestrated from here, but runs in HuggingFace / SageMaker / or elsewhere

    """

    def __init__(self, client: Client, config: Dict[str, Any] = None):
        super().__init__(client, config)

    def config_cls(self) -> Type[Config]:
        return TestConfig

    def model_cls(self) -> Type[TestTrainableTaggerConfigModel]:
        return TestTrainableTaggerConfigModel

    def run_with_model(
        self,
        request: PluginRequest[BlockAndTagPluginInput],
        model: TestTrainableTaggerConfigModel,
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Downloads the model file from the provided workspace"""
        logging.debug(f"run_with_model {request} {model}")
        logging.info(
            f"TestTrainableTaggerPlugin:run_with_model() got request {request} and model {model}"
        )
        return model.run(request)

    def get_training_parameters(
        self, request: PluginRequest[TrainingParameterPluginInput]
    ) -> InvocableResponse[TrainingParameterPluginOutput]:
        ret = InvocableResponse(data={})
        return ret

    def train(
        self, request: PluginRequest[TrainPluginInput], model: TestTrainableTaggerConfigModel
    ) -> InvocableResponse[TrainPluginOutput]:
        _ = model.train(request)
        return InvocableResponse(data=TrainPluginOutput())

    def train_status(
        self, request: PluginRequest[TrainPluginInput], model: TrainableModel
    ) -> InvocableResponse[TrainPluginOutput]:
        model.train_status(request)
        return InvocableResponse(data=TrainPluginOutput())


handler = create_handler(TestTrainableTaggerConfigPlugin)



from enum import Enum
from typing import Optional


class TagKind(str, Enum):
    """A set of `kind` constants for Tags.

    These define broad categories of tags. Suggested `name` values for each category are found in
    separate enums. For example: kind=TagKind.DOCUMENT, name=DocumentTag.H1
    """

    PART_OF_SPEECH = "part-of-speech"
    DEPENDENCY = "dependency"
    SENTIMENT = "sentiment"
    EMOTION = "emotion"
    ENTITY = "entity"
    DOCUMENT = "document"
    TOKEN = "token"  # noqa: S105
    INTENT = "intent"
    EMBEDDING = "embedding"


class DocTag(str, Enum):
    """A set of `name` constants for for Tags with a `kind` of `TagKind.doc`; appropriate for HTML and Markdown ideas."""

    DOCUMENT = "document"
    PAGE = "page"  # E.g. in a PDF
    REGION = "region"  # E.g., abstract catchall region in a document
    HEADER = "header"
    H1 = "h1"
    H2 = "h2"
    H3 = "h3"
    H4 = "h4"
    H5 = "h5"
    LINE = "line"
    TITLE = "title"
    SUBTITLE = "subtitle"
    FOOTER = "footer"
    PARAGRAPH = "paragraph"
    ORDERED_LIST = "ordered-list"
    UNORDERED_LIST = "unordered-list"
    LIST_ITEM = "list-item"
    LINK = "link"
    CAPTION = "caption"
    IMAGE = "image"
    BLOCK_QUOTE = "block-quote"
    BLOCK_CODE = "block-code"
    UNKNOWN = "unknown"
    SENTENCE = "sentence"
    TOKEN = "token"  # noqa: S105
    SPAN = "span"
    DIV = "div"
    PRE = "pre"
    STRONG = "strong"
    EMPHASIZED = "emphasized"
    UNDERLINED = "underlined"
    TELETYPE = "teletype"
    ARTICLE = "article"
    MAIN = "main"

    @staticmethod
    def from_html_tag(tagname: Optional[str]) -> Optional["DocTag"]:  # noqa: C901
        if tagname is None:
            return None

        name = tagname.lower().strip()

        if name == "p":
            return DocTag.PARAGRAPH
        elif name == "h1":
            return DocTag.H1
        elif name == "h2":
            return DocTag.H2
        elif name == "h3":
            return DocTag.H3
        elif name == "h4":
            return DocTag.H4
        elif name == "h5":
            return DocTag.H5
        elif name == "ul":
            return DocTag.UNORDERED_LIST
        elif name == "ol":
            return DocTag.ORDERED_LIST
        elif name == "li":
            return DocTag.LIST_ITEM
        elif name == "a":
            return DocTag.LINK
        elif name == "div":
            return DocTag.DIV
        elif name == "img":
            return DocTag.IMAGE
        elif name == "span":
            return DocTag.SPAN
        elif name == "pre":
            return DocTag.PRE
        elif name == "code":
            return DocTag.BLOCK_CODE
        elif name == "blockquote":
            return DocTag.BLOCK_QUOTE
        elif name == "strong":
            return DocTag.STRONG
        elif name == "b":
            return DocTag.STRONG
        elif name == "emph":
            return DocTag.EMPHASIZED
        elif name == "i":
            return DocTag.EMPHASIZED
        elif name == "u":
            return DocTag.UNDERLINED
        elif name == "tt":
            return DocTag.TELETYPE
        elif name == "article":
            return DocTag.ARTICLE
        elif name == "header":
            return DocTag.HEADER
        elif name == "footer":
            return DocTag.FOOTER
        elif name == "main":
            return DocTag.MAIN

        return None


class TokenTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.token`; appropriate for parsing-level ideas."""

    TEXT_WITH_WHITESPACE = "text-with-whitespace"
    WHITESPACE = "whitespace"
    HEAD = "head"
    LEFT_EDGE = "left-edge"
    RIGHT_EDGE = "right-edge"
    ENTITY_TYPE = "entity-type"
    ENTITY_IOB = "entity-iob"
    LEMMA = "lemma"
    NORMALIZED = "normalized"
    SHAPE = "shape"
    PREFIX = "prefix"
    SUFFIX = "suffix"
    IS_ALPHA = "is-alpha"
    IS_ASCII = "is-ascii"
    IS_DIGIT = "is-digit"
    IS_TITLE = "is-title"
    IS_PUNCT = "is-punct"
    IS_LEFT_PUNCT = "is-left-punct"
    IS_RIGHT_PUNCT = "is-right-punct"
    IS_SPACE = "is-space"
    IS_BRACKET = "is-bracket"
    IS_QUOTE = "is-quote"
    IS_CURRENCY = "is-currency"
    LIKE_URL = "like-url"
    LIKE_NUM = "like-num"
    LIKE_EMAIL = "like-email"
    IS_OUT_OF_VOCABULARY = "is-out-of-vocabulary"
    IS_STOPWORD = "is-stopword"
    LANGUAGE = "language"


class SentimentTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.sentiment`."""

    POSITIVE = "positive"
    NEUTRAL = "neutral"
    NEGATIVE = "negative"
    SCORE = "score"


class EmotionTag(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.emotion`."""

    POSITIVE = "positive"
    NEUTRAL = "neutral"
    NEGATIVE = "negative"
    HAPPINESS = "happiness"
    SADNESS = "sadness"
    JOY = "joy"
    LOVE = "love"
    ANGER = "anger"
    FEAR = "fear"
    SURPRISE = "surprise"
    HUMOR = "humor"
    CONCERN = "concern"
    SERIOUSNESS = "seriousness"
    SCORE = "score"


class IntentTags(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.intent`."""

    SALUTATION = "salutation"
    PRAISE = "praise"
    COMPLAINT = "complaint"
    QUESTION = "question"
    REQUEST = "request"
    EXPLANATION = "explanation"
    SCHEDULING_REQUEST = "scheduling-request"
    ARE_YOU_THERE = "are-you-there"
    REVISITING_TOPIC = "revisiting-topic"


class TagValue(str, Enum):
    """A set of key constants for the `value` object within a tag.."""

    SCORE = "score"
    VECTOR_VALUE = "vector-value"
    NUMBER_VALUE = "number-value"
    BOOL_VALUE = "bool-value"
    STRING_VALUE = "string-value"


class EntityTags(str, Enum):
    """A set of `name` constants for Tags with a `kind` of `TagKind.entity`."""

    PERSON = "person"
    ORGANIZATION = "organization"
    PRODUCT = "product"
    LOCATION = "location"
    DATE = "date"
    TIME = "time"
    MONEY = "money"
    PERCENT = "percent"
    FACILITY = "facility"
    GEO_POLITICAL_ENTITY = "geo-political-entity"



"""
Pytest reads this file in before doing any work. Read more about conftest.py under:
  - https://docs.pytest.org/en/stable/fixture.html
  - https://docs.pytest.org/en/stable/writing_plugins.html
"""

import sys
from pathlib import Path

# Make sure `steamship_tests` is on the PYTHONPATH. Otherwise cross-test imports (e.g. to util libraries) will fail.
sys.path.append(str(Path(__file__).parent.absolute()))

from steamship_tests.utils.fixtures import client, invocable_handler  # noqa: F401, E402



import pytest
from assets.plugins.blockifiers.blockifier import TEST_DOC
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin

from steamship import File
from steamship.client import Steamship


@pytest.mark.usefixtures("client")
def test_e2e_importer(client: Steamship):
    file_importer_path = PLUGINS_PATH / "importers" / "plugin_file_importer.py"
    with deploy_plugin(client, file_importer_path, "fileImporter") as (
        plugin,
        version,
        instance,
    ):
        # The test FileImporter should always return a string file with contents TEST_DOC
        file = File.create_with_plugin(client=client, plugin_instance=instance.handle)
        file.wait()
        file = file.output
        # Now fetch the data from Steamship and assert that it is the SAME as the data the FileImporter creates
        data = file.raw()
        assert data.decode("utf-8") == TEST_DOC

        file.delete()



from typing import Type

from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.file_importer import FileImporter
from steamship.plugin.inputs.file_import_plugin_input import FileImportPluginInput
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput


class TestFileImporterPythonErrorPlugin(FileImporter):
    class EmptyConfig(Config):
        pass

    def config_cls(self) -> Type[Config]:
        return self.EmptyConfig

    def run(
        self, request: PluginRequest[FileImportPluginInput]
    ) -> InvocableResponse[RawDataPluginOutput]:
        raise Exception("TestFileImporterPythonErrorPlugin")


handler = create_handler(TestFileImporterPythonErrorPlugin)



import pprint
from typing import Callable, Optional

import pytest
from assets.packages.demo_package import TestPackage


@pytest.mark.parametrize("invocable_handler", [TestPackage], indirect=True)
def test_package_spec(invocable_handler: Callable[[str, str, Optional[dict]], dict]):
    """Test that the handler returns the proper directory information"""
    response_dict = invocable_handler("GET", "/__dir__", {})
    pp = pprint.PrettyPrinter(indent=4)
    print()
    pp.pprint(response_dict)



from typing import Type

from steamship.invocable import Config, Invocable, InvocableResponse, create_handler, get, post


class HelloWorld(Invocable):
    def config_cls(self) -> Type[Config]:
        return Config

    @post("greet")
    def greet(self, name: str = "Person") -> InvocableResponse:
        return InvocableResponse(string=f"Hello, {name}")

    @get("workspace")
    def workspace(self) -> InvocableResponse:
        return InvocableResponse(string=self.client.config.workspace_id)


handler = create_handler(HelloWorld)



import pytest
from assets.plugins.blockifiers.blockifier import DummyBlockifierPlugin

from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput

TEST_REQ = RawDataPluginInput(data="Hi there")
TEST_PLUGIN_REQ = PluginRequest(data=TEST_REQ)
TEST_PLUGIN_REQ_DICT = TEST_PLUGIN_REQ.dict()


def _test_resp(res):
    assert res.data is not None
    assert res.data.file is not None
    assert len(res.data.file.blocks) == 4


def test_resp():
    blockifier = DummyBlockifierPlugin()
    request = PluginRequest(data=TEST_REQ)
    res = blockifier.run(request)
    _test_resp(res)

    # The endpoints take a kwargs block which is transformed into the appropriate JSON object
    res2 = blockifier.run_endpoint(**TEST_PLUGIN_REQ_DICT)
    _test_resp(res2)

    # This plugin is not trainable, and thus it refuses trainable parameters requests
    with pytest.raises(AttributeError):
        blockifier.get_training_parameters(PluginRequest(data=TrainingParameterPluginInput()))
    with pytest.raises(AttributeError):
        blockifier.get_training_parameters_endpoint(
            **PluginRequest(data=TrainingParameterPluginInput()).dict()
        )

    # This plugin is not trainable, and thus it refuses train requests
    with pytest.raises(AttributeError):
        blockifier.train(PluginRequest(data=TrainPluginInput()))
    with pytest.raises(AttributeError):
        blockifier.train_endpoint(**PluginRequest(data=TrainPluginInput()).dict())



from assets.packages.hello_world import HelloWorld, handler

from steamship.invocable import InvocableRequest, Invocation

NAME = "Ted"
RES_EMPTY = "Hello, Person"
RES_NAME = f"Hello, {NAME}"


def test_invoke_package_in_python():
    package = HelloWorld()

    assert package.greet().data == RES_EMPTY
    assert package.greet(NAME).data == RES_NAME


def test_invoke_package_with_request():
    package = HelloWorld()

    req = InvocableRequest(invocation=Invocation(http_verb="POST", invocation_path="greet"))
    res = package(req)
    assert res.data == RES_EMPTY

    req = InvocableRequest(
        invocation=Invocation(http_verb="POST", invocation_path="greet", arguments={"name": NAME})
    )
    res = package(req)
    assert res.data == RES_NAME


def test_invoke_package_with_handler():
    logging_config = {"loggingHost": "none", "loggingPort": "none"}
    event = {
        "invocation": {"httpVerb": "POST", "invocationPath": "greet"},
        "loggingConfig": logging_config,
        "invocationContext": {},
        "clientConfig": {
            "workspaceId": "dummy_workspace_id",
            "workspaceHandle": "dummy_workspace_handle",
        },
    }
    res = handler(event)
    assert res["data"] == RES_EMPTY

    event["invocation"]["arguments"] = {"name": NAME}
    res = handler(event)
    assert res["data"] == RES_NAME



from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from steamship import Configuration, PackageInstance, PluginInstance, SteamshipError, Workspace
from steamship.base.client import Client
from steamship.data.embeddings import EmbedAndSearchRequest, QueryResults

_logger = logging.getLogger(__name__)


class Steamship(Client):
    """Steamship Python Client."""

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        workspace: str = None,
        fail_if_workspace_exists: bool = False,
        profile: str = None,
        config_file: str = None,
        config: Configuration = None,
        trust_workspace_config: bool = False,  # For use by lambda_handler; don't fetch the workspace
        **kwargs,
    ):
        super().__init__(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            web_base=web_base,
            workspace=workspace,
            fail_if_workspace_exists=fail_if_workspace_exists,
            profile=profile,
            config_file=config_file,
            config=config,
            trust_workspace_config=trust_workspace_config,
            **kwargs,
        )
        # We use object.__setattr__ here in order to bypass Pydantic's overloading of it (which would block this
        # set unless we were to add this as a field)
        object.__setattr__(self, "use", self._instance_use)
        object.__setattr__(self, "use_plugin", self._instance_use_plugin)

    def __repr_args__(self: BaseModel) -> Any:
        """Because of the trick we've done with `use` and `use_plugin`, we need to exclude these from __repr__
        otherwise we'll get an infinite recursion."""
        return [
            (key, value)
            for key, value in self.__dict__.items()
            if key != "use" and key != "use_plugin"
        ]

    def embed_and_search(
        self,
        query: str,
        docs: List[str],
        plugin_instance: str,
        k: int = 1,
    ) -> QueryResults:
        req = EmbedAndSearchRequest(query=query, docs=docs, plugin_instance=plugin_instance, k=k)
        return self.post(
            "plugin/instance/embeddingSearch",
            req,
            expect=QueryResults,
        )

    @staticmethod
    def use(
        package_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        reuse: bool = True,
        workspace_handle: Optional[str] = None,
        **kwargs,
    ) -> PackageInstance:
        """Creates/loads an instance of package `package_handle`.

        The instance is named `instance_handle` and located in the Workspace named `instance_handle`.

        For example, one may write the following to always get back the same package instance, no matter how many
        times you run it, scoped into its own workspace:

        ```python
        plugin = Steamship.use('package-handle', 'instance-handle')
        ```

        If you wish to override the usage of a workspace named `instance_handle`, you can provide the `workspace_handle`
        parameter.
        """
        kwargs["workspace"] = workspace_handle or instance_handle
        client = Steamship(**kwargs)
        return client._instance_use(
            package_handle=package_handle,
            instance_handle=instance_handle,
            config=config,
            version=version,
            reuse=reuse,
        )

    def _instance_use(
        self,
        package_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        reuse: bool = True,
    ) -> PackageInstance:
        """Creates/loads an instance of package `package_handle`.

        The instance is named `instance_handle` and located in the workspace this client is anchored to.."""
        instance = PackageInstance.create(
            self,
            package_handle=package_handle,
            package_version_handle=version,
            handle=instance_handle,
            config=config,
            fetch_if_exists=reuse,
        )

        return instance

    @staticmethod
    def use_plugin(
        plugin_handle: str,
        instance_handle: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        reuse: bool = True,
        workspace_handle: Optional[str] = None,
        **kwargs,
    ) -> PluginInstance:
        """Creates/loads an instance of plugin `plugin_handle`.

        The instance is named `instance_handle` and located in the Workspace named `instance_handle`.

        For example, one may write the following to always get back the same plugin instance, no matter how many
        times you run it, scoped into its own workspace:

        ```python
        plugin = Steamship.use_plugin('plugin-handle', 'instance-handle')
        ```
        """
        kwargs["workspace"] = workspace_handle or instance_handle
        client = Steamship(**kwargs)
        return client._instance_use_plugin(
            plugin_handle=plugin_handle,
            instance_handle=instance_handle,
            config=config,
            version=version,
            reuse=reuse,
        )

    def _instance_use_plugin(
        self,
        plugin_handle: str,
        instance_handle: str = None,
        config: Dict[str, Any] = None,
        version: str = None,
        reuse: bool = True,
    ) -> PluginInstance:
        """Creates/loads an instance of plugin `plugin_handle`.

        The instance is named `instance_handle` and located in the workspace this client is anchored to."""
        instance = PluginInstance.create(
            self,
            plugin_handle=plugin_handle,
            plugin_version_handle=version,
            handle=instance_handle,
            config=config,
            fetch_if_exists=reuse,
        )

        return instance

    def get_workspace(self) -> Workspace:
        # We should probably add a hard-coded way to get this. The client in a Steamship Plugin/App comes
        # pre-configured with an API key and the Workspace in which this client should be operating.
        # This is a way to load the model object for that workspace.
        logging.info(
            f"get_workspace() called on client with config workspace {self.config.workspace_handle}/{self.config.workspace_id}"
        )
        workspace = Workspace.get(
            self, id_=self.config.workspace_id, handle=self.config.workspace_handle
        )
        if not workspace:
            logging.error("Unable to get workspace.")
            raise SteamshipError(
                message="Error while retrieving the Workspace associated with this client config.",
                internal_message=f"workspace_id={self.config.workspace_id} workspace_handle={self.config.workspace_handle}",
            )
        logging.info(f"Got workspace: {workspace.handle}/{workspace.id}")
        return workspace



from __future__ import annotations

from typing import Dict, Optional

from steamship.base.model import CamelModel
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput


class TrainingParameterPluginInput(CamelModel):
    # The plugin instance handle that should perform the training.
    plugin_instance: Optional[str] = None
    # An export request to produce the training data file, if training data is required.
    export_plugin_input: Optional[ExportPluginInput] = None

    # How many epochs to train (if supported by the supplied `pluginInstance`)
    training_epochs: Optional[int] = None

    # How much of the data to hold out for testing (if supported by the supplied `pluginInstance`)
    testing_holdout_percent: Optional[float] = None

    # Random seed for performing the train/test split (if supported by the supplied `pluginInstance`)
    test_split_seed: Optional[int] = None

    # Custom training-time parameters, specific to the pluginInstance
    training_params: Optional[Dict] = None

    # Custom inference-time parameters, specific to the pluginInstance
    inference_params: Optional[Dict] = None



from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import File


def test_e2e_blockifier_plugin():
    client = get_steamship_client()
    blockifier_path = PLUGINS_PATH / "blockifiers" / "blockifier.py"
    with deploy_plugin(client, blockifier_path, "blockifier") as (
        plugin,
        version,
        instance,
    ):
        file = File.create(client=client, content="This is a test.")
        assert len(file.refresh().blocks) == 0
        file.blockify(plugin_instance=instance.handle).wait()
        assert len(file.refresh().blocks) == 4
        file.delete()



from __future__ import annotations

from typing import Any, Dict, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import Request


class CreatePackageVersionRequest(Request):
    package_id: str = None
    handle: str = None
    type: str = "file"
    config_template: Dict[str, Any] = None


class PackageVersion(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    package_id: str = None
    handle: str = None
    config_template: Dict[str, Any] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["packageVersion"] if "packageVersion" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        package_id: str = None,
        handle: str = None,
        filename: str = None,
        filebytes: bytes = None,
        config_template: Dict[str, Any] = None,
    ) -> PackageVersion:

        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, "rb") as f:
                filebytes = f.read()

        req = CreatePackageVersionRequest(
            handle=handle, package_id=package_id, config_template=config_template
        )

        task = client.post(
            "package/version/create",
            payload=req,
            file=("package.zip", filebytes, "multipart/form-data"),
            expect=PackageVersion,
        )
        task.wait()
        return task.output



from typing import Type

from steamship import MimeTypes
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.file_importer import FileImporter
from steamship.plugin.inputs.file_import_plugin_input import FileImportPluginInput
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput

# Note: this aligns with the same document in the internal Engine test.
HANDLE = "test-importer-plugin-v1"
TEST_H1 = "A Poem"
TEST_S1 = "Roses are red."
TEST_S2 = "Violets are blue."
TEST_S3 = "Sugar is sweet, and I love you."
TEST_DOC = f"# {TEST_H1}\n\n{TEST_S1} {TEST_S2}\n\n{TEST_S3}\n"


class TestFileImporterPlugin(FileImporter):
    class EmptyConfig(Config):
        pass

    def config_cls(self) -> Type[Config]:
        return self.EmptyConfig

    def run(
        self, request: PluginRequest[FileImportPluginInput]
    ) -> InvocableResponse[RawDataPluginOutput]:
        bytes = (100000 * TEST_DOC).encode("utf-8")
        return InvocableResponse(data=RawDataPluginOutput(_bytes=bytes, mime_type=MimeTypes.MKD))


handler = create_handler(TestFileImporterPlugin)



from __future__ import annotations

import json
from typing import Any, Dict, List

from pydantic import Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import Request
from steamship.base.response import Response


class TagQueryRequest(Request):
    tag_filter_query: str


class Tag(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    file_id: str = None
    block_id: str = None
    kind: str = None  # E.g. ner
    name: str = None  # E.g. person
    value: Dict[str, Any] = None  # JSON Metadata
    start_idx: int = None  # w/r/t block.text. None means 0 if blockId is not None
    end_idx: int = None  # w/r/t block.text. None means -1 if blockId is not None

    class CreateRequest(Request):
        id: str = None
        file_id: str = None
        block_id: str = None
        kind: str = None
        name: str = None
        start_idx: int = None
        end_idx: int = None
        value: Dict[str, Any] = None

    class DeleteRequest(Request):
        id: str = None
        file_id: str = None
        block_id: str = None

    class ListRequest(Request):
        file_id: str = None
        block_id: str = None

    class ListResponse(Response):
        tags: List[Tag] = None

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        block_id: str = None,
        kind: str = None,
        name: str = None,
        start_idx: int = None,
        end_idx: int = None,
        value: Any = None,
    ) -> Tag:
        if isinstance(value, dict) or isinstance(value, list):
            value = json.dumps(value)

        req = Tag.CreateRequest(
            file_id=file_id,
            block_id=block_id,
            kind=kind,
            name=name,
            start_idx=start_idx,
            end_idx=end_idx,
            value=value,
        )
        return client.post("tag/create", req, expect=Tag)

    def delete(self) -> Tag:
        return self.client.post(
            "tag/delete",
            Tag.DeleteRequest(id=self.id, file_id=self.file_id, block_id=self.block_id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> TagQueryResponse:
        req = TagQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "tag/query",
            payload=req,
            expect=TagQueryResponse,
        )
        return res


class TagQueryResponse(Response):
    tags: List[Tag]


Tag.ListResponse.update_forward_refs()



from typing import Callable, Optional, Type

import pytest
from steamship_tests.utils.client import get_steamship_client
from steamship_tests.utils.random import random_name

from steamship import Steamship, Workspace
from steamship.invocable import InvocableRequest, Invocation, InvocationContext, LoggingConfig
from steamship.invocable.invocable import Invocable
from steamship.invocable.lambda_handler import create_handler as _create_handler


@pytest.fixture()
def client() -> Steamship:
    """Returns a client rooted in a new workspace, then deletes that workspace afterwards.

    To use, simply import this file and then write a test which takes `client`
    as an argument.

    Example
    -------
    The client can be used by injecting a fixture as follows::

        @pytest.mark.usefixtures("client")
        def test_something(client):
          pass
    """
    steamship = get_steamship_client()
    workspace = Workspace.create(client=steamship)
    new_client = get_steamship_client(workspace_id=workspace.id)
    yield new_client
    workspace.delete()


@pytest.fixture()
def invocable_handler(request) -> Callable[[str, str, Optional[dict]], dict]:
    """
    Returns a client rooted in a new workspace, then deletes that workspace afterwards.

    To use, simply import this file and then write a test which takes `invocable_handler`
    as an argument and parameterize it via PyTest.

    Example
    --------

        import pytest # doctest: +SKIP
        from steamship_tests.utils.fixtures import invocable_handler  # noqa: F401
        from assets.packages.demo_package import TestPackage

        @pytest.mark.parametrize("invocable_handler", [TestPackage], indirect=True)
            def _test_something(invocable_handler):
                response_dict = invocable_handler("POST", "/hello", dict())

    The invocable will be run its own workspace that gets cleaned up afterwards, and
    the test can be written from the perspective of an external caller of the
    invocable.
    """
    invocable: Type[Invocable] = request.param
    steamship = get_steamship_client()
    workspace_handle = random_name()
    workspace = Workspace.create(client=steamship, handle=workspace_handle)
    new_client = get_steamship_client(workspace=workspace_handle)

    def handle(verb: str, invocation_path: str, arguments: Optional[dict] = None) -> dict:
        _handler = _create_handler(invocable)
        invocation = Invocation(
            http_verb=verb, invocation_path=invocation_path, arguments=arguments or {}
        )
        logging_config = LoggingConfig(logging_host="none", logging_port="none")
        request = InvocableRequest(
            client_config=new_client.config,
            invocation=invocation,
            logging_config=logging_config,
            invocation_context=InvocationContext(),
        )
        event = request.dict(by_alias=True)
        return _handler(event)

    yield handle
    workspace.delete()



import pytest
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_index

from steamship import Steamship, SteamshipError
from steamship.base import TaskState
from steamship.data.embeddings import EmbeddedItem, EmbeddingIndex
from steamship.data.plugin.plugin_instance import PluginInstance

_TEST_EMBEDDER = "test-embedder"


def create_index(
    client: Steamship = None,
    plugin_instance: str = None,
    handle: str = None,
    fetch_if_exists: bool = None,
) -> EmbeddingIndex:
    client = client or get_steamship_client()
    return EmbeddingIndex.create(
        client=client,
        plugin_instance=plugin_instance,
        handle=handle,
        fetch_if_exists=fetch_if_exists,
    )


def test_create_index():
    client = get_steamship_client()
    plugin_instance = PluginInstance.create(client, plugin_handle=_TEST_EMBEDDER)

    index = create_index(
        client=client,
        plugin_instance=plugin_instance.handle,
    )

    assert index is not None

    # Duplicate creation should fail with fetch_if_exists=False
    with pytest.raises(SteamshipError):
        _ = create_index(
            client=client,
            handle=index.handle,
            plugin_instance=plugin_instance.handle,
            fetch_if_exists=False,
        )

    index.delete()


def test_delete_index():
    steamship = get_steamship_client()
    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    index = create_index(plugin_instance=plugin_instance.handle)
    assert index.id is not None

    index2 = create_index(handle=index.handle, plugin_instance=plugin_instance.handle)
    assert index.id == index2.id

    index.delete()

    index3 = create_index(plugin_instance=plugin_instance.handle)
    assert index.id != index3.id
    index3.delete()


def _list_equal(actual, expected):
    assert len(actual) == len(expected)
    assert all([a == b for a, b in zip(actual, expected)])


def test_insert_many():
    steamship = get_steamship_client()
    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance.handle) as index:
        item1 = EmbeddedItem(
            value="Pizza", external_id="pizza", external_type="food", metadata=[1, 2, 3]
        )
        item2 = EmbeddedItem(
            value="Rocket Ship",
            external_id="workspace",
            external_type="vehicle",
            metadata="Foo",
        )

        index.insert_many([item1, item2])
        index.embed().wait()

        list_response = index.list_items()
        assert len(list_response.items) == 2
        assert len(list_response.items[0].embedding) > 0
        assert len(list_response.items[1].embedding) > 0
        assert len(list_response.items[0].embedding) == len(list_response.items[1].embedding)

        res = index.search(item1.value, include_metadata=True, k=100)
        res.wait()
        items = res.output.items
        assert items is not None
        assert len(items) == 2
        assert items[0].value.value == item1.value
        assert items[0].value.external_id == item1.external_id
        assert items[0].value.external_type == item1.external_type
        _list_equal(items[0].value.metadata, item1.metadata)

        res = index.search(item2.value, include_metadata=True)
        res.wait()
        items = res.output.items
        assert items is not None
        assert items[0].value.value == item2.value
        assert items[0].value.external_id == item2.external_id
        assert items[0].value.external_type == item2.external_type
        assert items[0].value.metadata == item2.metadata


def test_embed_task():
    steamship = get_steamship_client()
    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance.handle) as index:
        _ = index.insert("test", reindex=False)
        res = index.embed()

        assert res.task_id is not None
        assert res.state is not None
        assert res.task_created_on is not None
        assert res.task_last_modified_on is not None
        assert res.state == TaskState.waiting
        res.wait()
        assert res.state == TaskState.succeeded


def test_duplicate_inserts():
    steamship = get_steamship_client()

    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance.handle) as index:
        # Test for suppressed re-indexing
        a1 = "Ted can eat an entire block of cheese."
        q1 = "Who can eat the most cheese"
        _ = index.insert(a1)
        _ = index.search(q1)


def test_index_usage():
    steamship = get_steamship_client()

    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance.handle) as index:
        a1 = "Ted can eat an entire block of cheese."
        q1 = "Who can eat the most cheese"
        _ = index.insert(a1)
        _ = index.search(q1)

        # Now embed
        task = index.embed()
        task.wait()
        assert task.state == TaskState.succeeded

        search_results = index.search(q1)
        search_results.wait()
        search_results = search_results.output.items
        assert len(search_results) == 1
        assert search_results[0].value.value == a1

        # Associate metadata
        a2 = "Armadillo shells are bulletproof."
        q2 = "What is something interesting about Armadillos?"
        a2id = "A2id"
        a2type = "A2type"
        a2metadata = {
            "id": a2id,
            "idid": f"{a2id}{a2id}",
            "boolVal": True,
            "intVal": 123,
            "floatVal": 1.2,
        }

        _ = index.insert(a2, external_id=a2id, external_type=a2type, metadata=a2metadata)
        search_results2 = index.search(q2)
        search_results2.wait()
        search_results = search_results2.output.items
        assert len(search_results) == 1
        assert search_results[0].value.value == a2
        assert search_results[0].value.external_id is None
        assert search_results[0].value.external_type is None
        assert search_results[0].value.metadata is None

        search_results3 = index.search(q2, include_metadata=True)
        search_results3.wait()
        search_results = search_results3.output.items
        assert len(search_results) == 1
        assert search_results[0].value.value == a2
        assert search_results[0].value.external_id == a2id
        assert search_results[0].value.external_type == a2type

        assert search_results[0].value.metadata == a2metadata
        # Because I don't know pytest enough to fully trust the dict comparison..
        assert search_results[0].value.metadata["id"] == a2id
        assert search_results[0].value.metadata["idid"] == f"{a2id}{a2id}"

        search_results4 = index.search(q2, k=10)
        search_results4.wait()
        search_results = search_results4.output.items
        assert len(search_results) == 2
        assert search_results[0].value.value == a2
        assert search_results[1].value.value == a1


def test_multiple_queries():
    steamship = get_steamship_client()

    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance.handle) as index:
        # Test for suppressed re-indexing
        a1 = "Ted can eat an entire block of cheese."
        a2 = "Joe can drink an entire glass of water."
        _ = index.insert_many([a1, a2])
        index.embed().wait()

        qs1 = ["Who can eat the most cheese", "Who can run the fastest?"]
        search_results = index.search(qs1)
        search_results.wait()
        search_results = search_results.output
        assert len(search_results.items) == 1
        assert search_results.items[0].value.value == a1
        assert search_results.items[0].value.query == qs1[0]

        qs2 = ["Who can tie a shoe?", "Who can drink the most water?"]
        search_results = index.search(qs2)
        search_results.wait()
        search_results = search_results.output
        assert len(search_results.items) == 1
        assert search_results.items[0].value.value == a2
        assert search_results.items[0].value.query == qs2[1]

        qs3 = ["What can Ted do?", "What can Sam do?", "What can Jerry do?"]
        search_results = index.search(qs3)
        search_results.wait()
        search_results = search_results.output
        assert len(search_results.items) == 1
        assert search_results.items[0].value.value == a1
        assert search_results.items[0].value.query == qs3[0]

        qs3 = ["What can Sam do?", "What can Ted do?", "What can Jerry do?"]
        search_results = index.search(qs3)
        search_results.wait()
        search_results = search_results.output
        assert len(search_results.items) == 1
        assert search_results.items[0].value.value == a1
        assert search_results.items[0].value.query == qs3[1]

        index.create_snapshot().wait()

        a3 = "Susan can run very fast."
        a4 = "Brenda can fight alligators."
        _ = index.insert_many([a3, a4])
        index.embed().wait()

        qs4 = ["What can Brenda do?", "What can Ronaldo do?", "What can Jerry do?"]
        search_results = index.search(qs4)
        search_results.wait()
        search_results = search_results.output
        assert len(search_results.items) == 1
        assert search_results.items[0].value.value == a4
        assert search_results.items[0].value.query == qs4[0]

        qs4 = [
            "What can Brenda do?",
            "Who should run a marathon?",
            "What can Jerry do?",
        ]
        search_results = index.search(qs4, k=2)
        search_results.wait()
        search_results = search_results.output
        assert len(search_results.items) == 2
        assert search_results.items[0].value.value == a4
        assert search_results.items[0].value.query == qs4[0]
        assert search_results.items[1].value.value == a3
        assert search_results.items[1].value.query == qs4[1]


def test_empty_queries():
    steamship = get_steamship_client()

    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance.handle) as index:
        a1 = "Ted can eat an entire block of cheese."
        a2 = "Joe can drink an entire glass of water."
        _ = index.insert_many([a1, a2])
        index.embed().wait()

        with pytest.raises(SteamshipError):
            _ = index.search(None)

        # These technically don't count as empty. Leaving this test in here
        # to encode and capture that in case we want to change it.
        search_results = index.search([])
        search_results.wait()
        search_results = search_results.output
        # noinspection PyUnresolvedReferences
        assert len(search_results.items) == 0

        search_results = index.search("")
        search_results.wait()
        search_results = search_results.output
        # noinspection PyUnresolvedReferences
        assert len(search_results.items) == 1



from .hosting import HostingCpu, HostingEnvironment, HostingMemory, HostingTimeout, HostingType
from .plugin import Plugin, PluginAdapterType, PluginTargetType, PluginType
from .plugin_instance import PluginInstance
from .plugin_version import PluginVersion

__all__ = [
    "Plugin",
    "PluginVersion",
    "PluginInstance",
    "HostingMemory",
    "HostingTimeout",
    "HostingCpu",
    "HostingEnvironment",
    "HostingType",
    "PluginType",
    "PluginAdapterType",
    "PluginTargetType",
]



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Dict, Generic, Optional, TypeVar

from typing_extensions import TypeAlias

from steamship.base.client import Client
from steamship.invocable import InvocableResponse
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.outputs.model_checkpoint import ModelCheckpoint
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput

ModelConstructor: TypeAlias = Callable[[], "TrainableModel"]

# Global variable to store the model for reuse in memory.
MODEL_CACHE: Dict[str, "TrainableModel"] = {}

ConfigType = TypeVar("ConfigType")


class TrainableModel(ABC, Generic[ConfigType]):
    """Base class for trainable models.

    Trainable models are not plugins. They are a thin wrapper around the state of a model designed to be **used with**
    the Steamship plugin system.

    # State Management

    100% of a TrainableModel's state management should save to & read from a folder on disk via the methods
    `save_to_folder` and `load_from_folder`.

    # Remote Saving and Loading

    `TrainableModel` instances automatically save to a user's Workspace on Steamship via `save_remote` method. They
    can load themselves from a user's workspace via the `load_remote` method.

    When saving a model, the caller provides `handle`, such as "V1" or "epoch_23". This allows that particular checkpoint
    to be re-loaded. By default, every save operation also saves the model to the "default" checkpoint, overwriting it
    if it already existed. When a user loads a model without specifying a checkpoint, the "default" checkpoint will be used.

    # Data Scope

    A TrainableModel's data is saved & loaded with respect to

    1) The user's active Workspace, and
    2) The provided Plugin Instance within that workspace.

    The active workspace is read from the Steamship client context, and the `plugin_instance_id` is supplied as a
    method argument on the `save_remote` and `load_remote` methods.

    This organization enables a user to have arbitrarily many trained model instances of the same type colocated within
    a Workspace.

    # Training

    A training job is fully parameterized by the `TrainPluginInput` object.

    # Result Reporting

    A training job's results are reported via the `TrainPluginOutput` object. These results include a reference to the
    `save_remote` output, but they do not include the model parameters themselves. For example, after training, one
    could write:

    >>> archive_path_in_steamship = model.save_remote(..)
    >>> output = TrainPluginOutput(archive_path_in_steamship=archive_path_in_steamship,
        ...
        )

    That output is the ultimate return object of the training process, but the Plugin that owns this model need not
    wait for synchronous completion to update the Steamship Engine with intermediate results. It can use the
    `Response.post_update` to proactively stream results back to the server.

    # Third-party / External Models

    This model class is a convenient wrapper for models running on third party systems (e.g. Google's AutoML). In such
    a case:

    - The `train` method would begin the job on the 3rd party system.
    - The `save_to_folder` method would write the Job ID and any other useful data to the checkpoint path
    - The `load_from_folder` method would read this Job ID from disk and obtain an authenticated client with the
      third party system.
    - Any `run` method the implementer created would ferry back results fetched from the third-party system.
    - Any status reporting in TrainPluginOutput would ferry back status fetched from the third-party system.

    """

    config: ConfigType = None

    def receive_config(self, config: ConfigType):
        """Stores config from plugin instance, so it is accessible by model on load or train."""
        self.config = config

    @abstractmethod
    def save_to_folder(self, checkpoint_path: Path):
        """Saves 100% of the state of this model to the provided path."""
        raise NotImplementedError()

    @abstractmethod
    def load_from_folder(self, checkpoint_path: Path):
        """Load 100% of the state of this model to the provided path."""
        raise NotImplementedError()

    @abstractmethod
    def train(self, input: PluginRequest[TrainPluginInput]) -> InvocableResponse[TrainPluginOutput]:
        """Train or fine-tune the model, parameterized by the information in the TrainPluginInput object."""
        raise NotImplementedError()

    @abstractmethod
    def train_status(
        self, input: PluginRequest[TrainPluginInput]
    ) -> InvocableResponse[TrainPluginOutput]:
        """Check on the status of an in-process training job, if it is running externally asynchronously."""
        raise NotImplementedError()

    @classmethod
    def load_from_local_checkpoint(cls, checkpoint: ModelCheckpoint, config: ConfigType):
        model = cls()
        model.receive_config(config=config)
        model.load_from_folder(checkpoint.folder_path_on_disk())
        return model

    @classmethod
    def load_remote(
        cls,
        client: Client,
        plugin_instance_id: str,
        checkpoint_handle: Optional[str] = None,
        use_cache: bool = True,
        model_parent_directory: Path = None,
        plugin_instance_config: ConfigType = None,
    ):
        if checkpoint_handle is None:
            # For some reason doing this defaulting in the signature wasn't working.
            checkpoint_handle = ModelCheckpoint.DEFAULT_HANDLE

        model_key = f"{plugin_instance_id}/{checkpoint_handle}"
        logging.info(f"TrainableModel:load_remote - Model Key: {model_key}")

        global MODEL_CACHE

        if use_cache:
            if model_key in MODEL_CACHE:
                logging.info(f"TrainableModel:load_remote - Returning cached: {model_key}")
                return MODEL_CACHE[model_key]

        checkpoint = ModelCheckpoint(
            client=client,
            parent_directory=model_parent_directory,
            handle=checkpoint_handle,
            plugin_instance_id=plugin_instance_id,
        )

        # If we haven't loaded the model, we need to download and start the model
        logging.info(f"TrainableModel:load_remote - Downloading: {model_key}")
        checkpoint.download_model_bundle()
        logging.info(f"TrainableModel:load_remote - Loading: {model_key}")
        model = cls.load_from_local_checkpoint(checkpoint, plugin_instance_config)
        logging.info(f"TrainableModel:load_remote - Loaded: {model_key}")

        if use_cache:
            MODEL_CACHE[model_key] = model

        return model

    def save_remote(
        self,
        client: Client,
        plugin_instance_id: str,
        checkpoint_handle: Optional[str] = None,
        model_parent_directory: Path = None,
        set_as_default: bool = True,
    ) -> str:
        if checkpoint_handle is None:
            # For some reason doing this defaulting in the signature wasn't working.
            checkpoint_handle = ModelCheckpoint.DEFAULT_HANDLE

        checkpoint = ModelCheckpoint(
            client=client,
            parent_directory=model_parent_directory,
            handle=checkpoint_handle,
            plugin_instance_id=plugin_instance_id,
        )
        self.save_to_folder(checkpoint.folder_path_on_disk())
        checkpoint.upload_model_bundle(set_as_default=set_as_default)
        return checkpoint.archive_path_in_steamship()



import os
import shutil
import tempfile
from pathlib import Path

from steamship_tests import TEST_ASSETS_PATH
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_name

from steamship import Workspace
from steamship.data.workspace import SignedUrl
from steamship.utils.signed_urls import download_from_signed_url, upload_to_signed_url
from steamship.utils.zip_archives import zip_folder


def test_upload_download_text():
    tempbase = Path(tempfile.mkdtemp())
    upfile = tempbase / "up.txt"
    downfile = tempbase / "down.txt"

    with open(upfile, "w") as f:
        f.write("Hello")

    # Grab a Steamship client and generate an upload url
    client = get_steamship_client()
    workspace = Workspace.get(client=client)
    upload_name = random_name()
    url_resp = workspace.create_signed_url(
        SignedUrl.Request(
            bucket=SignedUrl.Bucket.PLUGIN_DATA,
            filepath=upload_name,
            operation=SignedUrl.Operation.WRITE,
        )
    )
    assert url_resp is not None
    assert url_resp.signed_url is not None

    # Upload the zip file to the URL
    upload_to_signed_url(url_resp.signed_url, filepath=upfile)

    # Now create a download signed URL
    download_resp = workspace.create_signed_url(
        SignedUrl.Request(
            bucket=SignedUrl.Bucket.PLUGIN_DATA,
            filepath=upload_name,
            operation=SignedUrl.Operation.READ,
        )
    )
    assert download_resp is not None
    assert download_resp.signed_url is not None

    # Verify that we get an exception when downloading something that doesn't exist
    # TODO: Follow up after we get a firmer understaing of the failure semantics of Localstack 404 errors with
    # pre-signed URLs versus AWS Actual errors with pre-signed URLs. Localstack seems to reply with an HTTP 200
    # containing an XML string that contains an error message.

    # url = download_resp.data.signedUrl
    # parsed_url = urlparse(url)
    # bad_url = urlunparse(parsed_url._replace(path=f"{parsed_url.path}BOOGA"))
    # bad_download_path = tempbase / Path("bad.zip")
    # with pytest.raises(Exception):
    #     download_from_signed_url(download_resp.data.signedUrl, to_file=bad_download_path)

    # Download the zip file to the URL
    download_from_signed_url(download_resp.signed_url, to_file=downfile)

    # Verify the download URL is there
    assert os.path.exists(downfile)

    # Verify the zip files are the same
    with open(downfile, "r") as f1:
        with open(upfile, "r") as f2:
            f1c = f1.read()
            f2c = f2.read()
            assert f1c == f2c


def test_upload_download():
    """This test simulates some of the operations models must do when trainable/loading.

    It performs the following steps:

    1. Zips up a folder
    2. Creates a signed-url with Write access in the "model" bucket of a workspace
    3. Uploads the zip file
    4. Creates a signed-url with Read access in the "model" bucket of a workspace, pointing to the same file
    5. Downloads the file
    6. Verifies that they are the same file

    """
    # Copy the test assets to a temp folder
    tempbase = Path(tempfile.mkdtemp())
    shutil.copytree(TEST_ASSETS_PATH, os.path.join(tempbase, "src"))

    # Zip that folder
    zip_path = tempbase / Path("src.zip")
    zip_folder(tempbase / Path("src"), into_file=zip_path)

    # Verify that on disk, src.zip exists
    assert os.path.exists(zip_path)

    # Grab a Steamship client and generate an upload url
    client = get_steamship_client()
    workspace = Workspace.get(client=client)
    upload_name = random_name()
    url_resp = workspace.create_signed_url(
        SignedUrl.Request(
            bucket=SignedUrl.Bucket.PLUGIN_DATA,
            filepath=upload_name,
            operation=SignedUrl.Operation.WRITE,
        )
    )
    assert url_resp is not None
    assert url_resp.signed_url is not None

    # Upload the zip file to the URL
    upload_to_signed_url(url_resp.signed_url, filepath=zip_path)

    # Now create a download signed URL
    download_resp = workspace.create_signed_url(
        SignedUrl.Request(
            bucket=SignedUrl.Bucket.PLUGIN_DATA,
            filepath=upload_name,
            operation=SignedUrl.Operation.READ,
        )
    )
    assert download_resp is not None
    assert download_resp.signed_url is not None

    # Verify that we get an exception when downloading something that doesn't exist
    # TODO: Follow up after we get a firmer understaing of the failure semantics of Localstack 404 errors with
    # pre-signed URLs versus AWS Actual errors with pre-signed URLs. Localstack seems to reply with an HTTP 200
    # containing an XML string that contains an error message.

    # url = download_resp.data.signedUrl
    # parsed_url = urlparse(url)
    # bad_url = urlunparse(parsed_url._replace(path=f"{parsed_url.path}BOOGA"))
    # bad_download_path = tempbase / Path("bad.zip")
    # with pytest.raises(Exception):
    #     download_from_signed_url(download_resp.data.signedUrl, to_file=bad_download_path)

    # Download the zip file to the URL
    download_path = tempbase / Path("out.zip")
    download_from_signed_url(download_resp.signed_url, to_file=download_path)

    # Verify the download URL is there
    assert os.path.exists(download_path)

    # Verify the zip files are the same
    with open(download_path, "rb") as f1:
        with open(zip_path, "rb") as f2:
            f1c = f1.read()
            f2c = f2.read()
            assert f1c == f2c



__copyright__ = "Steamship"
__license__ = "MIT"

from unittest.mock import patch

import pytest
from pydantic import ValidationError
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Steamship
from steamship.base.client import Client
from steamship.base.configuration import DEFAULT_API_BASE, DEFAULT_APP_BASE, DEFAULT_WEB_BASE
from steamship.data.user import User


def test_get_steamship_client():
    client = get_steamship_client()
    assert client.config is not None
    assert client.config.profile == "test"
    assert client.config.api_key is not None
    user = User.current(client)
    assert user.id is not None
    assert user.handle is not None


TEST_WEB_BASE = "https://app.test.com/"
TEST_APP_BASE = "https://test.run/"
TEST_API_BASE = "https://api.test.com/api/v1/"

empty_base_uris = [
    (TEST_WEB_BASE, TEST_APP_BASE, TEST_API_BASE),
    (None, TEST_APP_BASE, TEST_API_BASE),
    (TEST_WEB_BASE, None, TEST_API_BASE),
    (TEST_WEB_BASE, TEST_APP_BASE, None),
]


def switch_workspace(
    self,
    workspace_handle: str = None,
    workspace_id: str = None,
    fail_if_workspace_exists: bool = False,
    trust_workspace_config: bool = False,
):
    pass


@pytest.mark.parametrize(("web_base", "app_base", "api_base"), empty_base_uris)
def test_base_uris(web_base: str, app_base: str, api_base: str) -> None:
    with patch.object(Client, "switch_workspace", switch_workspace):
        client = Steamship(api_base=api_base, web_base=web_base, app_base=app_base)
        assert str(client.config.web_base) == web_base or DEFAULT_WEB_BASE
        assert str(client.config.app_base) == app_base or DEFAULT_APP_BASE
        assert str(client.config.api_base) == api_base or DEFAULT_API_BASE


def test_incorrectly_formatted_base_uris() -> None:
    with patch.object(Client, "switch_workspace", switch_workspace):
        client = Steamship(
            api_base=TEST_API_BASE[:-1], web_base=TEST_WEB_BASE[:-1], app_base=TEST_APP_BASE[:-1]
        )
        assert str(client.config.web_base) == TEST_WEB_BASE
        assert str(client.config.app_base) == TEST_APP_BASE
        assert str(client.config.api_base) == TEST_API_BASE


@pytest.mark.parametrize("base_uri", ["", "test", "ftp://test.com"])
def test_invalid_base_uris(base_uri: str) -> None:
    with pytest.raises(ValidationError):
        Steamship(api_base=base_uri, web_base=base_uri, app_base=base_uri)


def test_empty_config() -> None:
    client = Steamship()
    assert client.config is not None
    assert client.config.web_base is not None
    assert client.config.app_base is not None
    assert client.config.api_base is not None


def test_none_config() -> None:
    client = Steamship(config=None)
    assert client.config is not None
    assert client.config.web_base is not None
    assert client.config.app_base is not None
    assert client.config.api_base is not None


class TestObject:
    pass


def test_incorrect_config_type() -> None:
    with pytest.raises(ValidationError):
        Steamship(config=TestObject())


@pytest.mark.parametrize(
    ("app_base", "user", "fixed_base"),
    [
        pytest.param("http://test:8081", "user", "http://test:8081/"),
        pytest.param("http://test:8081/", "user", "http://test:8081/"),
        pytest.param("http://localhost:8081", "user", "http://localhost:8081/"),
        pytest.param("http://localhost:8081/", "user", "http://localhost:8081/"),
        pytest.param("http://127.0.0.1:8081", "user", "http://127.0.0.1:8081/"),
        pytest.param("http://127.0.0.1:8081/", "user", "http://127.0.0.1:8081/"),
        pytest.param(
            "http://host.docker.internal:8081", "user", "http://host.docker.internal:8081/"
        ),
        pytest.param(
            "http://host.docker.internal:8081/", "user", "http://host.docker.internal:8081/"
        ),
        pytest.param("http://0:0:0:0:8081", "user", "http://0:0:0:0:8081/"),
        pytest.param("http://0:0:0:0:8081/", "user", "http://0:0:0:0:8081/"),
        pytest.param("https://apps.steamship.run", "user", "https://user.apps.steamship.run/"),
        pytest.param("https://apps.steamship.run/", "user", "https://user.apps.steamship.run/"),
        pytest.param(
            "https://apps.staging.steamship.com", "user", "https://user.apps.staging.steamship.com/"
        ),
        pytest.param(
            "https://apps.staging.steamship.com/",
            "user",
            "https://user.apps.staging.steamship.com/",
        ),
    ],
)
def test_invocable_call_rewriting(app_base: str, user: str, fixed_base: str):
    client = get_steamship_client()
    client.config.app_base = app_base
    operation = "foo"

    output_url = client._url(is_package_call=True, package_owner=user, operation=operation)
    assert output_url == f"{fixed_base}{operation}"



import os
import shutil
import tempfile
from pathlib import Path

from steamship_tests import TEST_ASSETS_PATH

from steamship.utils.zip_archives import unzip_folder, zip_folder


def test_zip_unzip():
    # Copy the test assets to a temp folder
    tempbase = tempfile.mkdtemp()
    shutil.copytree(TEST_ASSETS_PATH, os.path.join(tempbase, "src"))

    # Zip that folder
    zip_file = Path(tempbase) / Path("src.zip")
    zip_folder(Path(tempbase) / Path("src"), into_file=zip_file)

    # Verify that on disk, src.zip exists
    assert os.path.exists(os.path.join(tempbase, "src.zip"))

    # Copy it to dest.zip
    dest_path = Path(tempbase) / Path("dest.zip")
    shutil.move(zip_file, dest_path)

    # Verify that on disk, src.zip doesn't exist and dest.zip does
    assert not os.path.exists(os.path.join(tempbase, "src.zip"))
    assert os.path.exists(os.path.join(tempbase, "dest.zip"))
    assert not os.path.exists(os.path.join(tempbase, "dest"))

    # Unzip that folder
    dest_folder = Path(tempbase) / Path("dest")
    unzip_folder(dest_path, into_folder=dest_folder)

    # Verify that on disk, dest/ exists
    assert os.path.exists(dest_folder) is True
    assert os.path.isdir(dest_folder) is True

    # Verify that the contents of dest are the contents of src
    src_files = os.listdir(Path(tempbase) / Path("src"))
    dest_files = os.listdir(dest_folder)
    assert src_files == dest_files



import json
import logging
import sys
import uuid
from http import HTTPStatus
from typing import Dict, Type

from fluent import asynchandler as fluenthandler
from fluent.handler import FluentRecordFormatter

from steamship import Configuration
from steamship.base import SteamshipError
from steamship.client import Steamship
from steamship.data.workspace import SignedUrl
from steamship.invocable import InvocableRequest, InvocableResponse, InvocationContext
from steamship.invocable.invocable import Invocable
from steamship.utils.signed_urls import upload_to_signed_url


def encode_exception(obj):
    """When logging an exception ex: logging.exception(some_error), the exception must be turned into a string
    so that it is accepted by elasticsearch"""
    if isinstance(obj, SteamshipError):
        return json.dumps(obj.to_dict())
    if isinstance(obj, Exception):
        return f"exception_class: {type(obj).__name__}, args: {obj.args}"
    return obj


def create_handler(invocable_cls: Type[Invocable]):  # noqa: C901
    """Wrapper function for a Steamship invocable within an AWS Lambda function."""

    def _handler(
        event: Dict,
        client: Steamship,
        _: Dict = None,
    ) -> InvocableResponse:

        try:
            request = InvocableRequest.parse_obj(event)
        except SteamshipError as se:
            logging.exception(se)
            return InvocableResponse.from_obj(se)
        except Exception as ex:
            logging.exception(ex)
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message="Plugin/App handler was unable to parse inbound request.",
                exception=ex,
            )

        if request and request.invocation:
            error_prefix = (
                f"[ERROR - {request.invocation.http_verb} {request.invocation.invocation_path}] "
            )
        else:
            error_prefix = "[ERROR - ?VERB ?PATH] "

        try:
            invocable = invocable_cls(client=client, config=request.invocation.config)
        except SteamshipError as se:
            return InvocableResponse.from_obj(se)
        except Exception as ex:
            logging.exception(ex)
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                prefix=error_prefix,
                message="Unable to initialize package/plugin.",
                exception=ex,
            )

        if not invocable:
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                prefix=error_prefix,
                message="Unable to construct package/plugin for invocation.",
            )

        try:
            response = invocable(request)
            return InvocableResponse.from_obj(response)
        except SteamshipError as se:
            logging.exception(se)
            se.message = f"{error_prefix}{se.message}"
            return InvocableResponse.from_obj(se)
        except Exception as ex:
            logging.exception(ex)
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                prefix=error_prefix,
                exception=ex,
            )

    def handler(event: Dict, context: Dict = None) -> dict:
        logging_config = event.get("loggingConfig")

        if logging_config is None:
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message="Plugin/App handler did not receive a remote logging config.",
            ).dict(by_alias=True)

        logging_host = logging_config.get("loggingHost")
        logging_port = logging_config.get("loggingPort")

        logging.basicConfig(level=logging.INFO)
        logging_handler = None

        invocation_context_dict = event.get("invocationContext")
        if invocation_context_dict is None:
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message="Plugin/App handler did not receive an invocation context.",
            ).dict(by_alias=True)

        invocation_context = InvocationContext.parse_obj(invocation_context_dict)
        # These log statements intentionally go to the logging handler pre-remote attachment, to debug logging configuration issues
        logging.info(f"Logging host: {logging_host} Logging port: {logging_port}")
        logging.info(f"Invocation context: {invocation_context}")

        if (
            logging_host != "none"
        ):  # Key off the string none, not 'is None', to avoid config errors where remote host isn't passed
            # Configure remote logging
            if logging_host is None:
                return InvocableResponse.error(
                    code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    message="Plugin/App handler did receive a remote logging config, but it did not include a loggingHost.",
                ).dict(by_alias=True)

            if logging_port is None:
                return InvocableResponse.error(
                    code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    message="Plugin/App handler did receive a remote logging config, but it did not include a loggingPort.",
                ).dict(by_alias=True)

            custom_format = {
                "level": "%(levelname)s",
                "host": "%(hostname)s",
                "where": "%(module)s.%(filename)s.%(funcName)s:%(lineno)s",
                "type": "%(levelname)s",
                "stack_trace": "%(exc_text)s",
                "component": "package-plugin-lambda",
                "userId": invocation_context.user_id,
                "workspaceId": invocation_context.workspace_id,
                "tenantId": invocation_context.tenant_id,
                "invocableHandle": invocation_context.invocable_handle,
                "invocableVersionHandle": invocation_context.invocable_version_handle,
                "invocableType": invocation_context.invocable_type,
                "path": event.get("invocation", {}).get("invocationPath"),
            }
            logging_handler = fluenthandler.FluentHandler(
                "steamship.deployed_lambda",
                host=logging_host,
                port=logging_port,
                nanosecond_precision=True,
                msgpack_kwargs={"default": encode_exception},
            )
            formatter = FluentRecordFormatter(custom_format)
            logging_handler.setFormatter(formatter)
            # The below should make it so calls to logging.info etc are also routed to the remote logger
            logging.root.addHandler(logging_handler)

        try:
            # Config will accept `workspace_id` as passed from the Steamship Engine, whereas the `Steamship`
            # class itself is limited to accepting `workspace` (`config.workspace_handle`) since that is the manner
            # of interaction ideal for developers.
            config = Configuration(**event.get("clientConfig", {}))
            client = Steamship(config=config, trust_workspace_config=True)
        except SteamshipError as se:
            logging.exception(se)
            return InvocableResponse.from_obj(se).dict(by_alias=True)
        except Exception as ex:
            logging.exception(ex)
            return InvocableResponse.error(
                code=HTTPStatus.INTERNAL_SERVER_ERROR,
                message="Plugin/App handler was unable to create Steamship client.",
                exception=ex,
            ).dict(by_alias=True)

        response = _handler(event, client, context)

        result = response.dict(by_alias=True, exclude={"client"})
        # When created with data > 4MB, data is uploaded to a bucket.
        # This is a very ugly way to get the deep size of this object
        data = json.dumps(result.get("data", None)).encode("UTF-8")
        data_size = sys.getsizeof(data)
        logging.info(f"Response data size {data_size}")
        if data_size > 4e6 and invocation_context.invocable_type == "plugin":
            logging.info("Response data size >4MB, must upload to bucket")

            filepath = str(uuid.uuid4())
            signed_url = (
                client.get_workspace()
                .create_signed_url(
                    SignedUrl.Request(
                        bucket=SignedUrl.Bucket.PLUGIN_DATA,
                        filepath=filepath,
                        operation=SignedUrl.Operation.WRITE,
                    )
                )
                .signed_url
            )

            logging.info(f"Got signed url for writing: {signed_url}")

            upload_to_signed_url(signed_url, data)

            # Now remove raw data and replace with bucket
            del result["data"]
            result["dataBucket"] = SignedUrl.Bucket.PLUGIN_DATA.value
            result["dataFilepath"] = filepath

        if logging_handler is not None:
            logging_handler.close()

        logging.info(result)
        return result

    return handler



import pytest

from steamship import Block, File, MimeTypes
from steamship.client import Steamship
from steamship.data.tags import Tag


@pytest.mark.usefixtures("client")
def test_file_tag(client: Steamship):
    a = File.create(
        client,
        content="A",
        mime_type=MimeTypes.MKD,
    )
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD

    _ = Tag.create(client, file_id=a.id, kind="test1")
    _ = Tag.create(client, file_id=a.id, kind="test2")

    tags = Tag.query(client, tag_filter_query=f'file_id "{a.id}"')
    assert tags.tags is not None
    assert len(tags.tags) == 2

    must = ["test1", "test2"]
    for tag in tags.tags:
        assert tag.kind in must
        must.remove(tag.kind)
    assert len(must) == 0

    for tag in tags.tags:
        if tag.kind == "test1":
            tag.delete()

    Tag.create(client, file_id=a.id, kind="test3")

    tags = Tag.query(client, tag_filter_query=f'file_id "{a.id}"')
    assert tags is not None
    assert tags.tags is not None
    assert len(tags.tags) == 2

    must = ["test2", "test3"]
    for tag in tags.tags:
        assert tag.kind in must
        must.remove(tag.kind)
    assert len(must) == 0

    tags.tags[0].delete()
    tags.tags[1].delete()

    tags = Tag.query(client, tag_filter_query=f'file_id "{a.id}"')
    assert tags is not None
    assert tags.tags is not None
    assert len(tags.tags) == 0

    a.delete()


def test_query(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(kind="BlockTag")]),
            Block.CreateRequest(text="B"),
        ],
    )
    assert a.id is not None
    a = a.refresh()
    b = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A"),
            Block.CreateRequest(text="B", tags=[Tag.CreateRequest(kind="Test")]),
        ],
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    assert b.id is not None
    b = b.refresh()

    tags = Tag.query(client=client, tag_filter_query='blocktag and kind "BlockTag"').tags
    assert len(tags) == 1
    assert tags[0].id == a.blocks[0].tags[0].id

    tags = Tag.query(client=client, tag_filter_query='blocktag and kind "Test"').tags
    assert len(tags) == 1
    assert tags[0].id == b.blocks[1].tags[0].id

    a.delete()
    b.delete()



import json
import logging
from pathlib import Path

import pytest
from assets.plugins.taggers.plugin_trainable_tagger import TestTrainableTaggerModel
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.file import upload_file

from steamship.client import Steamship
from steamship.data.plugin import HostingType
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.model_checkpoint import ModelCheckpoint

EXPORTER_HANDLE = "signed-url-exporter"
KEYWORDS = ["product", "coupon"]


@pytest.mark.usefixtures("client")
def test_e2e_trainable_tagger_lambda_training(client: Steamship):
    version_config_template = {
        "text_column": {"type": "string"},
        "tag_columns": {"type": "string"},
        "tag_kind": {"type": "string"},
    }
    instance_config = {"text_column": "Message", "tag_columns": "Category", "tag_kind": "Intent"}

    exporter_plugin_r = PluginInstance.create(
        client=client,
        handle=EXPORTER_HANDLE,
        plugin_handle=EXPORTER_HANDLE,
        fetch_if_exists=True,
    )
    assert exporter_plugin_r is not None
    exporter_plugin = exporter_plugin_r
    assert exporter_plugin.handle is not None

    csv_blockifier_path = PLUGINS_PATH / "blockifiers" / "csv_blockifier.py"
    trainable_tagger_path = PLUGINS_PATH / "taggers" / "plugin_trainable_tagger.py"

    # Make a blockifier which will generate our trainable corpus
    with deploy_plugin(
        client,
        csv_blockifier_path,
        "blockifier",
        version_config_template=version_config_template,
        instance_config=instance_config,
    ) as (plugin, version, instance):
        with upload_file(client, "utterances.csv") as file:
            assert len(file.refresh().blocks) == 0
            # Use the plugin we just registered
            file.blockify(plugin_instance=instance.handle).wait()
            assert len(file.refresh().blocks) == 5

            # Now make a trainable tagger to train on those tags
            with deploy_plugin(
                client, trainable_tagger_path, "tagger", training_platform=HostingType.LAMBDA
            ) as (tagger, tagger_version, tagger_instance):
                # Now train the plugin
                training_request = TrainingParameterPluginInput(
                    plugin_instance=tagger_instance.handle,
                    export_plugin_input=ExportPluginInput(
                        plugin_instance=EXPORTER_HANDLE, type="file", query='kind "foo1"'
                    ),
                    training_params={
                        "keyword_list": KEYWORDS  # This is a key defined by the test model we're training
                    },
                )

                train_result = tagger_instance.train(training_request)
                train_result.wait()

                # At this point, the PluginInstance will have written a parameter file to disk. We should be able to
                # retrieve it since we know that it is tagged as the `default`.

                checkpoint = ModelCheckpoint(
                    client=client,
                    handle="default",
                    plugin_instance_id=tagger_instance.id,
                )
                checkpoint_path = checkpoint.download_model_bundle()
                assert checkpoint_path.exists()
                keyword_path = Path(checkpoint_path) / TestTrainableTaggerModel.KEYWORD_LIST_FILE
                assert keyword_path.exists()
                with open(keyword_path, "r") as f:
                    params = json.loads(f.read())
                    assert params == KEYWORDS

                logging.info("Waiting 15 seconds for instance to deploy.")
                import time

                time.sleep(15)

                # If we're here, we have verified that the plugin instance has correctly recorded its parameters
                # into the pluginData bucket under a path unique to the PluginInstnace/ModelCheckpoint.

                # Now we'll attempt to USE this plugin. This plugin's behavior is to simply tag any file with the
                # tags that parameter it. Since those tags are (see above) ["product", "coupon"] we should expect
                # this tagger to apply those tags to any file provided to it.

                # First we'll create a file
                test_doc = "Hi there"
                res = tagger_instance.tag(doc=test_doc)
                res.wait()
                assert res.output is not None
                assert res.output.file is not None
                assert res.output.file.tags is not None
                assert len(res.output.file.tags) == len(KEYWORDS)
                assert sorted([tag.kind for tag in res.output.file.tags]) == sorted(KEYWORDS)



from .tag import Tag
from .tag_constants import DocTag, EmotionTag, SentimentTag, TagKind, TagValue, TokenTag

__all__ = ["DocTag", "EmotionTag", "SentimentTag", "Tag", "TagKind", "TokenTag", "TagValue"]



from __future__ import annotations

from steamship.base.model import CamelModel


class TrainStatusPluginInput(CamelModel):
    """
    This is the object passed as input to a train_status operation, stored as the `input` field of a `train_status` task.
    """

    # Arbitrary key-valued data to provide keys into an async training process, assuming it is not complete
    # It will be passed back to the plugin in subsequent train_status calls
    training_reference_data: dict = None



from contextlib import contextmanager
from typing import Any, Dict

from steamship import Configuration, Steamship
from steamship.data.package.package_instance import PackageInstance
from steamship.data.plugin.plugin_instance import PluginInstance


def get_steamship_client(fail_if_workspace_exists=False, **kwargs) -> Steamship:
    # Always use the `test` profile
    kwargs["profile"] = "test"
    return Steamship(
        fail_if_workspace_exists=fail_if_workspace_exists, config=Configuration.parse_obj(kwargs)
    )


@contextmanager
def steamship_use(
    package_handle: str,
    instance_handle: str = None,
    config: Dict[str, Any] = None,
    version: str = None,
    reuse: bool = True,
    delete_workspace: bool = True,
    **kwargs
) -> PackageInstance:
    # Always use the `test` profile
    kwargs["profile"] = "test"
    instance = Steamship.use(package_handle, instance_handle, config, version, reuse, **kwargs)
    assert instance.client.config.workspace_id == instance.workspace_id
    yield instance
    # Clean up the workspace
    if delete_workspace:
        instance.client.get_workspace().delete()


@contextmanager
def steamship_use_plugin(
    plugin_handle: str,
    instance_handle: str = None,
    config: Dict[str, Any] = None,
    version: str = None,
    reuse: bool = True,
    delete_workspace: bool = True,
    **kwargs
) -> PluginInstance:
    # Always use the `test` profile
    kwargs["profile"] = "test"
    instance = Steamship.use_plugin(
        plugin_handle, instance_handle, config, version, reuse, **kwargs
    )
    assert instance.client.config.workspace_id == instance.workspace_id
    yield instance
    # Clean up the workspace
    if delete_workspace:
        instance.client.get_workspace().delete()



from .configuration import Configuration
from .error import SteamshipError
from .mime_types import MimeTypes
from .tasks import Task, TaskState

__all__ = ["Configuration", "SteamshipError", "Task", "TaskState", "MimeTypes"]



# TODO: It should fail if the docs field is empty.
# TODO: It should fail if the file hasn't been converted.
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_index

from steamship import File, MimeTypes, PluginInstance

_TEST_EMBEDDER = "test-embedder"

T = "A nice poem"
P1_1 = "Roses are red."
P1_2 = "Violets are blue."
P2_1 = "Sugar is sweet."
P2_2 = "I love you."
T2 = "A flavorful story"
P3_1 = "Cake is made of flour."
P3_2 = "Cake tastes good with milk."
P4_1 = "Cake comes in chocolate and vanilla flavors."
P4_2 = "Cake can be cut into mAny pieces and shared."


def test_file_parse():
    steamship = get_steamship_client()
    content1 = f"# {T}\n\n{P1_1} {P1_2}\n\n{P2_1} {P2_2}"
    content2 = f"# {T2}\n\n{P3_1} {P3_2}\n\n{P4_1} {P4_2}"
    content = f"{content1}\n\n{content2}"

    file = File.create(
        steamship,
        content=content,
        mime_type=MimeTypes.MKD,
    )
    assert file.id is not None
    assert file.mime_type == MimeTypes.MKD

    blockify_resp = file.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_resp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_resp = file.tag(plugin_instance=parser.handle)
    parse_resp.wait()

    # Now the sentences should be parsed!
    q2 = file.refresh()
    assert len(q2.blocks) == 6

    # Now we add the file to the index
    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    with random_index(steamship, plugin_instance=plugin_instance.handle) as index:
        index.insert_file(file.id, reindex=False)
        embed_resp = index.embed()
        embed_resp.wait()

        res = index.search("What color are roses?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
        assert items[0].value.value == " ".join([P1_1, P1_2])

    file.delete()


def test_file_index():
    steamship = get_steamship_client()
    t = "A nice poem"
    p1_1 = "Roses are red."
    p1_2 = "Violets are blue."
    p2_1 = "Sugar is sweet."
    p2_2 = "I love you."
    t2 = "A flavorful story"
    p3_1 = "Cake is made of flour."
    p3_2 = "Cake tastes good with milk."
    p4_1 = "Cake comes in chocolate and vanilla flavors."
    p4_2 = "Cake can be cut into mAny pieces and shared."

    content1 = f"# {t}\n\n{p1_1} {p1_2}\n\n{p2_1} {p2_2}"
    content2 = f"# {t2}\n\n{p3_1} {p3_2}\n\n{p4_1} {p4_2}"
    content = f"{content1}\n\n{content2}"

    file = File.create(
        steamship,
        content=content,
        mime_type=MimeTypes.MKD,
    )
    assert file.id is not None
    assert file.mime_type == MimeTypes.MKD

    blockify_resp = file.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_resp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_resp = file.tag(plugin_instance=parser.handle)
    parse_resp.wait()

    # Now the sentences should be parsed!
    q2 = file.refresh()
    assert len(q2.blocks) == 6

    # Now we add the file to the index via the shortcut.
    embedder = PluginInstance.create(steamship, plugin_handle="test-embedder")
    # noinspection PyUnresolvedReferences
    index = file.index(plugin_instance=embedder.handle)

    res = index.search("What color are roses?")
    res.wait()
    items = res.output.items
    assert len(items) == 1
    # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
    assert items[0].value.value == " ".join([p1_1, p1_2])

    res = index.search("What flavors does cake come in?")
    res.wait()
    items = res.output.items
    assert len(items) == 1
    # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
    assert items[0].value.value == " ".join([p4_1, p4_2])

    index.delete()
    file.delete()


def test_file_embed_lookup():
    steamship = get_steamship_client()

    content_a = "Ted likes to run."
    content_b = "Grace likes to bike."

    file = File.create(
        steamship,
        content=content_a,
        mime_type=MimeTypes.MKD,
    )

    blockify_task = file.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_task.wait()

    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_res = file.tag(plugin_instance=parser.handle)
    parse_res.wait()

    b = File.create(client=steamship, content=content_b, mime_type=MimeTypes.MKD)
    blockify_task = b.blockify(plugin_instance="markdown-blockifier-default-1.0")
    blockify_task.wait()

    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    parse_res = b.tag(plugin_instance=parser.handle)
    parse_res.wait()

    embedder = PluginInstance.create(steamship, plugin_handle="test-embedder")
    # Now we add the file to the index
    with random_index(steamship, embedder.handle) as index:
        index.insert_file(file.id, block_type="sentence", reindex=True)
        index.insert_file(b.id, block_type="sentence", reindex=True)

        res = index.search("What does Ted like to do?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        assert items[0].value.value == content_a

        res = index.search("What does Grace like to do?")
        res.wait()
        items = res.output.items
        assert len(items) == 1
        assert items[0].value.value == content_b

        # Now we list the items
        itemsa = index.list_items(file_id=file.id)
        assert len(itemsa.items) == 1
        assert len(itemsa.items[0].embedding) > 0
        assert itemsa.items[0].value == content_a

        itemsb = index.list_items(file_id=b.id)
        assert len(itemsb.items) == 1
        assert len(itemsb.items[0].embedding) > 0
        assert len(itemsb.items[0].embedding) == len(itemsa.items[0].embedding)
        assert itemsb.items[0].value == content_b



import base64
import json
import re

import pytest
import requests
from steamship_tests import PACKAGES_PATH, TEST_ASSETS_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import PackageInstance, SteamshipError, Workspace
from steamship.base import TaskState
from steamship.base.mime_types import MimeTypes
from steamship.utils.url import Verb


def _fix_url(s: str) -> str:
    """Homogenize references to `this machine` for the purpose of comparing remote configuration and local
    configuration. The goal of the below steamship_tests isn't to check that your machine has been configured in the
    "approved way" (which is to use host.docker.internal). It is merely to make sure that the environment
    has been passed to the invocable instance correctly."""
    s = s.replace("localhost", "host.docker.internal").replace("127.0.0.1", "host.docker.internal")
    if s.endswith("/"):
        s = s[:-1]  # s.removesuffix is only available in Python 3.9; we use Python 3.8
    return s


def test_instance_invoke():
    palm_tree_path = TEST_ASSETS_PATH / "palm_tree.png"

    with palm_tree_path.open("rb") as f:
        palm_bytes = f.read()
    base64_palm = base64.b64encode(palm_bytes).decode("utf-8")

    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "demo_package.py"

    with deploy_package(client, demo_package_path) as (package, version, instance):
        # Now let's invoke it!
        # Note: we're invoking the data at demo_package.py in the tests/assets/packages folder

        def get_raw(path: str):
            return requests.get(
                instance.full_url_for(path),
                headers={"authorization": f"Bearer {client.config.api_key}"},
            )

        res = instance.invoke("greet", verb=Verb.GET)
        assert res == "Hello, Person!"

        resp = get_raw("greet")
        assert resp.text == "Hello, Person!"

        res = instance.invoke("greet", verb=Verb.GET, name="Ted")
        assert res == "Hello, Ted!"
        url = instance.full_url_for("greet?name=Ted")
        resp = requests.get(url, headers={"authorization": f"Bearer {client.config.api_key}"})
        assert resp.text == "Hello, Ted!"

        res = instance.invoke("greet", verb=Verb.POST)
        assert res == "Hello, Person!"
        url = instance.full_url_for("greet")
        resp = requests.post(url, headers={"authorization": f"Bearer {client.config.api_key}"})
        assert resp.text == "Hello, Person!"

        res = instance.invoke("greet", verb=Verb.POST, name="Ted")
        assert res == "Hello, Ted!"
        url = instance.full_url_for("greet")
        resp = requests.post(
            url,
            json={"name": "Ted"},
            headers={"authorization": f"Bearer {client.config.api_key}"},
        )
        assert resp.text == "Hello, Ted!"

        # Now we test different return types
        resp_string = get_raw("resp_string")
        assert resp_string.text == "A String"

        resp_dict = get_raw("resp_dict")
        assert resp_dict.json() == {"string": "A String", "int": 10}

        resp_404 = get_raw("doesnt_exist")
        json_404 = resp_404.json()
        assert isinstance(json_404, dict)
        assert json_404.get("status") is not None
        assert json_404.get("status") is not None
        assert json_404.get("status", {}).get("state") == TaskState.failed
        # assert "No handler" in json_404.get("status", dict()).get("statusMessage", "")
        assert resp_404.status_code == 404

        resp_obj = get_raw("resp_obj")
        assert resp_obj.json() == {"name": "Foo"}

        resp_binary = get_raw("resp_binary")
        base64_binary = base64.b64encode(resp_binary.content).decode("utf-8")
        assert base64_binary == base64_palm
        assert resp_binary.headers.get("Content-Type") == MimeTypes.BINARY

        resp_bytes_io = get_raw("resp_bytes_io")
        base64_bytes_io = base64.b64encode(resp_bytes_io.content).decode("utf-8")
        assert base64_bytes_io == base64_palm
        assert resp_bytes_io.headers.get("Content-Type") == MimeTypes.BINARY

        resp_image = get_raw("resp_image")
        base64_image = base64.b64encode(resp_image.content).decode("utf-8")
        assert base64_image == base64_palm
        assert resp_image.headers.get("Content-Type") == MimeTypes.PNG

        # The test invocable, when executing remotely inside Steamship, should have the same
        # set of configuration options that we're running with here within the test
        configuration_within_lambda = instance.invoke("config", verb=Verb.GET)

        my_app_base = _fix_url(client.config.app_base)
        remote_app_base = _fix_url(configuration_within_lambda["appBase"])

        my_api_base = _fix_url(client.config.api_base)
        remote_api_base = _fix_url(configuration_within_lambda["apiBase"])

        assert my_app_base == remote_app_base
        assert my_api_base == remote_api_base

        # API key should NOT be the same as the original, because the invocable should be given a workspace-scoped key
        assert configuration_within_lambda["apiKey"] != client.config.api_key

        # WorkspaceId is an exception. Rather than being the WorkspaceId of the client, it should be the WorkspaceId
        # of the App Instance.
        assert configuration_within_lambda["workspaceId"] == instance.workspace_id  # WorkspaceID

        # The test invocable should NOT be able to fetch the User's account info.
        with pytest.raises(SteamshipError) as excinfo:
            _ = instance.invoke("user_info", verb=Verb.POST)
        assert "Cannot use a workspace-scoped key" in str(excinfo.value)

        # Test a JSON response that contains {"status": "a string"} in it to make sure the client base
        # isn't trying to coerce it to a Task object and throwing.
        resp_obj = instance.invoke("json_with_status", verb=Verb.POST)
        assert resp_obj == {"status": "a string"}

        # Test that the __steamship_dir__ method works
        #
        # Note: The output of the InvocableResponse includes a dynamically generated value:
        #
        #       returns': "<class 'l_dd6730cb.InvocableResponse[bytes]'>"
        #
        # So we've got to do a little regex switcheroo

        search_pattern = r"'[a-zA-Z0-9_]+\.InvocableResponse\["
        replace_pattern = "'foo.InvocableResponse["

        def replace_string(s: str):
            return re.sub(search_pattern, replace_pattern, s)

        steamship_dir = instance.invoke("__dir__")
        with open(TEST_ASSETS_PATH / "demo_package_spec.json", "r") as f:
            steamship_dir_golden = json.loads(replace_string(f.read()))
            steamship_dir_fixed = json.loads(replace_string(json.dumps(steamship_dir)))
            assert steamship_dir_fixed == steamship_dir_golden


def test_deploy_in_workspace():
    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "demo_package.py"

    workspace = Workspace.create(client)
    client.switch_workspace(workspace_id=workspace.id)

    assert workspace.handle != "default"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        # The Engine believes the instance to be in the workspace
        assert instance.workspace_id == workspace.id

        # The invocable believes itself to be in the workspace
        configuration_within_lambda = instance.invoke("config", verb=Verb.GET)
        assert configuration_within_lambda["workspaceId"] == workspace.id

    workspace.delete()


def test_package_instance_get():
    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "demo_package.py"

    workspace = Workspace.create(client)
    client.switch_workspace(workspace_id=workspace.id)

    assert workspace.handle != "default"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        instance_handle = instance.handle
        other_instance = PackageInstance.get(client, instance_handle)
        assert other_instance.id == instance.id
        assert other_instance.handle == instance.handle
        assert other_instance.package_id == instance.package_id
        assert other_instance.package_version_id == instance.package_version_id



from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client


def test_e2e_tagger():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "taggers" / "plugin_logging_tagger.py"
    with deploy_plugin(client, parser_path, "tagger") as (plugin, version, instance):
        test_doc = "Hi there"
        res = instance.tag(doc=test_doc)
        res.wait()
        assert res.output is not None
        assert len(res.output.file.blocks) == 1
        assert res.output.file.blocks[0].text == test_doc



from __future__ import annotations

import json
import logging
import typing
from abc import ABC
from inspect import isclass
from typing import Any, Dict, List, Type, TypeVar, Union

import inflection
from pydantic import BaseModel, PrivateAttr
from requests import Session

from steamship.base.configuration import Configuration
from steamship.base.error import SteamshipError
from steamship.base.mime_types import MimeTypes
from steamship.base.model import CamelModel, to_camel
from steamship.base.request import Request
from steamship.base.tasks import Task, TaskState
from steamship.utils.url import Verb, is_local

_logger = logging.getLogger(__name__)

T = TypeVar("T")  # TODO (enias): Do we need this?


class Client(CamelModel, ABC):
    """Client model.py class.

    Separated primarily as a hack to prevent circular imports.
    """

    config: Configuration
    _session: Session = PrivateAttr()

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        workspace: str = None,
        fail_if_workspace_exists: bool = False,
        profile: str = None,
        config_file: str = None,
        config: Configuration = None,
        trust_workspace_config: bool = False,  # For use by lambda_handler; don't fetch the workspace
        **kwargs,
    ):
        """Create a new client.

        If `workspace` is provided, it will anchor the client in a workspace by that name, creating it if necessary.
        Otherwise the `default` workspace will be used.
        """
        if config is not None and not isinstance(config, Configuration):
            config = Configuration.parse_obj(config)

        config = config or Configuration(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            web_base=web_base,
            workspace_handle=workspace,
            profile=profile,
            config_file=config_file,
        )
        self._session = Session()
        super().__init__(config=config)
        # The lambda_handler will pass in the workspace via the workspace_id, so we need to plumb this through to make sure
        # that the workspace switch performed doesn't mistake `workspace=None` as a request for the default workspace
        self.switch_workspace(
            workspace_handle=workspace or config.workspace_handle,
            workspace_id=config.workspace_id,
            fail_if_workspace_exists=fail_if_workspace_exists,
            trust_workspace_config=trust_workspace_config,
        )

    def switch_workspace(  # noqa: C901
        self,
        workspace_handle: str = None,
        workspace_id: str = None,
        fail_if_workspace_exists: bool = False,
        trust_workspace_config: bool = False,
        # For use by lambda_handler; don't fetch the workspacetrust_workspace_config: bool = False, # For use by lambda_handler; don't fetch the workspace
    ):
        """Switches this client to the requested workspace, possibly creating it. If all arguments are None, the client
        actively switches into the default workspace.

        - API calls are performed manually to not result in circular imports.
        - Note that the default workspace is technically not necessary for API usage; it will be assumed by the Engine
          in the absense of a Workspace ID or Handle being manually specified in request headers.
        """
        workspace = None

        if workspace_handle is None and workspace_id is None:
            # Switch to the default workspace since no named or ID'ed workspace was provided
            workspace_handle = "default"

        if fail_if_workspace_exists:
            logging.info(
                f"[Client] Creating workspace with handle/id: {workspace_handle}/{workspace_id}."
            )
        else:
            logging.info(
                f"[Client] Creating/Fetching workspace with handle/id: {workspace_handle}/{workspace_id}."
            )

        # Zero out the workspace_handle on the config block in case we're being invoked from
        # `init` (otherwise we'll attempt to create the sapce IN that nonexistant workspace)
        old_workspace_handle = self.config.workspace_handle
        self.config.workspace_handle = None

        if trust_workspace_config:
            if workspace_handle is None or workspace_id is None:
                raise SteamshipError(
                    message="Attempted a trusted workspace switch without providing both workspace handle and workspace id."
                )
            return_id = workspace_id
            return_handle = workspace_handle
        else:
            try:
                if workspace_handle is not None and workspace_id is not None:
                    get_params = {
                        "handle": workspace_handle,
                        "id": workspace_id,
                        "fetchIfExists": False,
                    }
                    workspace = self.post("workspace/get", get_params)
                elif workspace_handle is not None:
                    get_params = {
                        "handle": workspace_handle,
                        "fetchIfExists": not fail_if_workspace_exists,
                    }
                    workspace = self.post("workspace/create", get_params)
                elif workspace_id is not None:
                    get_params = {"id": workspace_id}
                    workspace = self.post("workspace/get", get_params)

            except SteamshipError as e:
                self.config.workspace_handle = old_workspace_handle
                raise e

            if workspace is None:
                raise SteamshipError(
                    message="Was unable to switch to new workspace: server returned empty Workspace."
                )

            return_id = workspace.get("workspace", {}).get("id")
            return_handle = workspace.get("workspace", {}).get("handle")

        if return_id is None or return_handle is None:
            raise SteamshipError(
                message="Was unable to switch to new workspace: server returned empty ID and Handle."
            )

        # Finally, set the new workspace.
        self.config.workspace_id = return_id
        self.config.workspace_handle = return_handle
        logging.info(f"[Client] Switched to workspace {return_handle}/{return_id}")

    def dict(self, **kwargs) -> dict:
        # Because of the trick we do to hack these in as both static and member methods (with different
        # implementations), Pydantic will try to include them by default. So we have to suppress that otherwise
        # downstream serialization into JSON will fail.
        if "exclude" not in kwargs:
            kwargs["exclude"] = {"use", "use_plugin", "_instance_use", "_instance_use_plugin"}
        elif isinstance(kwargs["exclude"], set):
            kwargs["exclude"].add("use")
            kwargs["exclude"].add("use_plugin")
            kwargs["exclude"].add("_instance_use")
            kwargs["exclude"].add("_instance_use_plugin")
        return super().dict(**kwargs)

    def _url(
        self,
        is_package_call: bool = False,
        package_owner: str = None,
        operation: str = None,
    ):
        if not is_package_call:
            # Regular API call
            base = self.config.api_base
        else:
            # Do the invocable version
            if package_owner is None:
                return SteamshipError(
                    code="UserMissing",
                    message="Cannot invoke a package endpoint without the package owner's user handle.",
                    suggestion="Provide the package_owner option, or initialize your package with a lookup.",
                )

            base = self.config.app_base
            if not is_local(base):
                # We want to prepend the user handle
                parts = base.split("//")
                base = f"{parts[0]}//{package_owner}.{'//'.join(parts[1:])}"

        # Clean leading and trailing "/"
        if base[len(base) - 1] == "/":
            base = base[:-1]
        if operation[0] == "/":
            operation = operation[1:]

        return f"{base}/{operation}"

    def _headers(
        self,
        is_package_call: bool = False,
        package_owner: str = None,
        package_id: str = None,
        package_instance_id: str = None,
        as_background_task: bool = False,
        wait_on_tasks: List[Task] = None,
    ):
        headers = {"Authorization": f"Bearer {self.config.api_key}"}

        if self.config.workspace_id:
            headers["X-Workspace-Id"] = self.config.workspace_id
        elif self.config.workspace_handle:
            headers["X-Workspace-Handle"] = self.config.workspace_handle

        if is_package_call:
            if package_owner:
                headers["X-Package-Owner-Handle"] = package_owner
            if package_id:
                headers["X-Package-Id"] = package_id
            if package_instance_id:
                headers["X-Package-Instance-Id"] = package_instance_id

        if wait_on_tasks:
            # Will result in the engine persisting the inbound HTTP request as a Task for deferred
            # execution. Additionally, the task will be scheduled to first wait on the other tasks
            # provided in the list of IDs.
            as_background_task = True
            headers["X-Task-Dependency"] = ",".join([task.task_id for task in wait_on_tasks])

        if as_background_task:
            # Will result in the engine persisting the inbound HTTP request as a Task for deferred
            # execution. The client will receive task information back instead of the synchronous API response.
            # That task can be polled for eventual completion.
            headers["X-Task-Background"] = "true"

        return headers

    @staticmethod
    def _prepare_data(payload: Union[Request, dict]):
        if payload is None:
            data = {}
        elif isinstance(payload, dict):
            data = payload
        elif isinstance(payload, BaseModel):
            data = payload.dict(by_alias=True)
        else:
            raise RuntimeError(f"Unable to parse payload of type {type(payload)}")

        return data

    @staticmethod
    def _response_data(resp, raw_response: bool = False):
        if resp is None:
            return None

        if raw_response:
            return resp.content

        if resp.headers:
            ct = None
            if "Content-Type" in resp.headers:
                ct = resp.headers["Content-Type"]
            if "content-type" in resp.headers:
                ct = resp.headers["content-type"]
            if ct is not None:
                ct = ct.split(";")[0]  # application/json; charset=utf-8
                if ct in [MimeTypes.TXT, MimeTypes.MKD, MimeTypes.HTML]:
                    return resp.text
                elif ct == MimeTypes.JSON:
                    return resp.json()
                else:
                    return resp.content

    @staticmethod
    def _prepare_multipart_data(data, file):
        # Note: requests seems to have a bug passing boolean (and maybe numeric?)
        # values in the midst of multipart form data. You need to manually convert
        # it to a string; otherwise it will pass as False or True (with the capital),
        # which is not standard notation outside of Python.
        for key in data:
            if data[key] is False:
                data[key] = "false"
            elif data[key] is True:
                data[key] = "true"

        result = {}
        for key, val in data.items():
            if val:
                if isinstance(val, Dict):
                    result[key] = (None, json.dumps(val), "application/json")
                else:
                    result[key] = (None, str(val))
        result["file"] = file
        return result

    def _add_client_to_response(self, expect: Type, response_data: Any):
        if isinstance(response_data, dict):
            self._add_client_to_object(expect, response_data)
        elif isinstance(response_data, list):
            for el in response_data:
                typing_parameters = typing.get_args(expect)
                self._add_client_to_response(
                    typing_parameters[0] if typing_parameters else None, el
                )

        return response_data

    def _add_client_to_object(self, expect, response_data):
        if expect and isclass(expect):
            if len(response_data.keys()) == 1 and list(response_data.keys())[0] in (
                to_camel(expect.__name__),
                to_camel(expect.__name__).replace("package", "invocable"),
                # Hack since engine uses "App" instead of "Package"
                "index",
            ):
                # TODO (enias): Hack since the engine responds with incosistent formats e.g. {"plugin" : {plugin_fields}}
                for _, v in response_data.items():
                    self._add_client_to_response(expect, v)
            elif issubclass(expect, BaseModel):
                response_data["client"] = self
                try:
                    key_to_type = typing.get_type_hints(expect)
                    for k, v in response_data.items():
                        self._add_client_to_response(key_to_type.get(inflection.underscore(k)), v)
                except NameError:
                    # typing.get_type_hints fails for Workspace
                    pass

    def call(  # noqa: C901
        self,
        verb: Verb,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Type[T] = None,
        debug: bool = False,
        raw_response: bool = False,
        is_package_call: bool = False,
        package_owner: str = None,
        package_id: str = None,
        package_instance_id: str = None,
        as_background_task: bool = False,
        wait_on_tasks: List[Task] = None,
    ) -> Union[
        Any, Task
    ]:  # TODO (enias): I would like to list all possible return types using interfaces instead of Any
        """Post to the Steamship API.

        All responses have the format::

        .. code-block:: json

        {
            "data": "<actual response>",
            "error": {"reason": "<message>"}
        } # noqa: RST203

        For the Python client we return the contents of the `data` field if present, and we raise an exception
        if the `error` field is filled in.
        """
        # TODO (enias): Review this codebase
        url = self._url(
            is_package_call=is_package_call,
            package_owner=package_owner,
            operation=operation,
        )

        headers = self._headers(
            is_package_call=is_package_call,
            package_owner=package_owner,
            package_id=package_id,
            package_instance_id=package_instance_id,
            as_background_task=as_background_task,
            wait_on_tasks=wait_on_tasks,
        )

        data = self._prepare_data(payload=payload)

        logging.info(
            f"Making {verb} to {url} in workspace {self.config.workspace_handle}/{self.config.workspace_id}"
        )
        if verb == Verb.POST:
            if file is not None:
                files = self._prepare_multipart_data(data, file)
                resp = self._session.post(url, files=files, headers=headers)
            else:
                resp = self._session.post(url, json=data, headers=headers)
        elif verb == Verb.GET:
            resp = self._session.get(url, params=data, headers=headers)
        else:
            raise Exception(f"Unsupported verb: {verb}")

        logging.info(f"From {verb} to {url} got HTTP {resp.status_code}")

        if debug is True:
            logging.debug(f"Got response {resp}")

        response_data = self._response_data(resp, raw_response=raw_response)

        logging.debug(f"Response JSON {response_data}")

        task = None
        error = None

        if isinstance(response_data, dict):
            if "status" in response_data:
                try:
                    task = Task.parse_obj(
                        {**response_data["status"], "client": self, "expect": expect}
                    )
                    if "state" in response_data["status"]:
                        if response_data["status"]["state"] == "failed":
                            error = SteamshipError.from_dict(response_data["status"])
                            logging.warning(f"Client received error from server: {error}")
                except TypeError as e:
                    # There's an edge case here -- if a Steamship package returns the JSON dictionary
                    #
                    # { "status": "status string" }
                    #
                    # Then the above handler will attempt to parse it and throw... But we don't actually want to throw
                    # since we don't take a strong opinion on what the response type of a package endpoint ought to be.
                    # It *may* choose to conform to the SteamshipResponse<T> type, but it doesn't have to.
                    if not is_package_call:
                        raise e

                if task is not None and task.state == TaskState.failed:
                    error = task.as_error()

            if "data" in response_data:
                if expect is not None:
                    if issubclass(expect, SteamshipError):
                        data = expect.from_dict({**response_data["data"], "client": self})
                    elif issubclass(expect, BaseModel):
                        data = expect.parse_obj(
                            self._add_client_to_response(expect, response_data["data"])
                        )
                    else:
                        raise RuntimeError(f"obj of type {expect} does not have a from_dict method")
                else:
                    data = response_data["data"]

                if task:
                    task.output = data
            else:
                data = response_data

        else:
            data = response_data

        if error is not None:
            logging.warning(f"Client received error from server: {error}", exc_info=error)
            raise error

        elif task is not None:
            return task
        elif data is not None:
            return data
        else:
            raise SteamshipError("Inconsistent response from server. Please contact support.")

    def post(
        self,
        operation: str,
        payload: Union[Request, dict, BaseModel] = None,
        file: Any = None,
        expect: Any = None,
        debug: bool = False,
        raw_response: bool = False,
        is_package_call: bool = False,
        package_owner: str = None,
        package_id: str = None,
        package_instance_id: str = None,
        as_background_task: bool = False,
        wait_on_tasks: List[Task] = None,
    ) -> Union[
        Any, Task
    ]:  # TODO (enias): I would like to list all possible return types using interfaces instead of Any
        return self.call(
            verb=Verb.POST,
            operation=operation,
            payload=payload,
            file=file,
            expect=expect,
            debug=debug,
            raw_response=raw_response,
            is_package_call=is_package_call,
            package_owner=package_owner,
            package_id=package_id,
            package_instance_id=package_instance_id,
            as_background_task=as_background_task,
            wait_on_tasks=wait_on_tasks,
        )

    def get(
        self,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Any = None,
        debug: bool = False,
        raw_response: bool = False,
        is_package_call: bool = False,
        package_owner: str = None,
        package_id: str = None,
        package_instance_id: str = None,
        as_background_task: bool = False,
        wait_on_tasks: List[Task] = None,
    ) -> Union[
        Any, Task
    ]:  # TODO (enias): I would like to list all possible return types using interfaces instead of Any
        return self.call(
            verb=Verb.GET,
            operation=operation,
            payload=payload,
            file=file,
            expect=expect,
            debug=debug,
            raw_response=raw_response,
            is_package_call=is_package_call,
            package_owner=package_owner,
            package_id=package_id,
            package_instance_id=package_instance_id,
            as_background_task=as_background_task,
            wait_on_tasks=wait_on_tasks,
        )



#
# This is the CLIENT-side abstraction for an invocable.
#
# If you are implementing a package, see: steamship.invocable.server.App
#

from __future__ import annotations

from typing import Any, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.request import CreateRequest, GetRequest


class Package(BaseModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["package"] if "package" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(client: Client, handle: str = None) -> Package:
        req = CreateRequest(handle=handle)
        return client.post("package/create", payload=req, expect=Package)

    @staticmethod
    def get(client: Client, handle: str) -> Package:
        return client.post("package/get", GetRequest(handle=handle), expect=Package)



import pytest
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError, Workspace


def test_default_workspace():
    client = get_steamship_client()
    workspace = Workspace.get(client=client)
    assert workspace is not None
    assert workspace.handle == "default"


def test_delete_workspace():
    client = get_steamship_client()
    default = Workspace.get(client=client)
    workspace1 = Workspace.create(client=client, handle="test")
    assert default.id is not None
    assert workspace1.id is not None
    assert default.id != workspace1.id

    workspace1.delete()
    with pytest.raises(SteamshipError):
        _ = Workspace.get(client=client, id_=workspace1.id)

    workspace1 = Workspace.create(client=client, handle="test")
    assert default.id is not None
    assert workspace1.id is not None
    assert default.id != workspace1.id

    workspace1a = Workspace.get(client=client, id_=workspace1.id)
    assert workspace1a is not None

    workspace1.delete()
    with pytest.raises(SteamshipError):
        _ = Workspace.get(client=client, id_=workspace1.id)


def test_get_workspace():
    client = get_steamship_client()
    default = Workspace.get(client=client)
    workspace1 = Workspace.create(client=client, handle="test")
    workspace1a = Workspace.get(client=client, id_=workspace1.id)
    assert workspace1a.id == workspace1.id
    assert workspace1a.id != default.id
    assert workspace1a.handle == workspace1.handle

    workspace1b = Workspace.get(client=client, handle=workspace1.handle)
    assert workspace1b.id == workspace1.id
    assert workspace1b.handle == workspace1.handle

    workspace1c = Workspace.get(client=client, id_=workspace1.id, handle=workspace1.handle)
    assert workspace1c.id == workspace1.id
    assert workspace1c.handle == workspace1.handle

    workspace1.delete()


def test_create_use_delete_workspace():
    client = get_steamship_client()
    default = Workspace.get(client=client)
    workspace1 = Workspace.create(client=client, handle="test")
    workspace2 = Workspace.create(client=client, handle="test2")

    assert workspace1 is not None
    assert workspace1.handle == "test"

    assert workspace2 is not None
    assert workspace2.handle == "test2"

    assert workspace2.id != workspace1.id
    assert workspace1.id != default.id
    assert workspace2.id != default.id

    workspace1a = Workspace.get(client=client, id_=workspace1.id)
    workspace1b = Workspace.get(client=client, handle=workspace1.handle)
    workspace1c = Workspace.get(client=client, id_=workspace1.id)

    assert workspace1.id == workspace1a.id
    assert workspace1.id == workspace1b.id
    assert workspace1.id == workspace1c.id

    workspace1a.delete()

    # These two are the same workspace! You can't delete twice!
    with pytest.raises(SteamshipError):
        workspace1b.delete()

    with pytest.raises(SteamshipError):
        workspace1c.delete()

    workspace2.delete()
    with pytest.raises(SteamshipError):
        _ = Workspace.get(client=client, id_=workspace1.id)


def test_list_workspace():
    client = get_steamship_client()
    default = Workspace.get(client=client)

    initial_workspace_count = len(Workspace.list(client).workspaces)

    workspace1 = Workspace.create(client=client)
    workspace2 = Workspace.create(client=client)
    workspace3 = Workspace.create(client=client)

    workspaces = Workspace.list(client).workspaces
    assert len(workspaces) == initial_workspace_count + 3
    assert default in workspaces
    assert workspace1 in workspaces
    assert workspace2 in workspaces
    assert workspace3 in workspaces

    workspace1.delete()
    workspace2.delete()
    workspace3.delete()



import re
from typing import TypeVar

import inflection
from pydantic import BaseModel
from pydantic.generics import GenericModel

T = TypeVar("T")  # Declare type variable


def to_camel(s: str) -> str:
    s = re.sub("_(url)$", lambda m: f"_{m.group(1).upper()}", s)
    return inflection.camelize(s, uppercase_first_letter=False)


class CamelModel(BaseModel):
    def __init__(self, **kwargs):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        super().__init__(**kwargs)

    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True
        # Populate enum values with their value, rather than the raw enum. Important to serialise model.dict()
        use_enum_values = True


class GenericCamelModel(CamelModel, GenericModel):
    pass



"""A simple key-value store implemented atop Files and Tags."""

from typing import Any, Dict, Optional

from steamship import Block, File, Steamship, Tag


class KeyValueStore:
    """A simple key value store implemented in Steamship.

    Instances of the KeyValueStore are identified by its  `namespace`.
    This store_identifier corresponds to a File that will be created with a special tag identifying it.

    Entries of the KeyValueStore are saved as `Tag` objects with:
      * Kind = "KeyValueStore"
      * Name = the key of the (kv) pair
      * Value = a dict set to the value

    Note that the value is always saved as a dict object. To save a string or int, wrap it in a dict.

    WARNING:

    This is essentially a clever hack atop Steamship's tag system to provide mutable key-value storage. It is in the
    steamship.utils package because it's proven useful once or twice. But in general, if you find yourself heavily
    relying upon it, consider reaching out to us at hello@steamship.com to let us know, and we'll up-prioritize
    adding a proper key-value API.
    """

    client: Steamship
    store_identifier: str

    def __init__(self, client: Steamship, store_identifier: str = "KeyValueStore"):
        """Create a new KeyValueStore instance.

        Args:
            client (Steamship): The Steamship client.
            store_identifier (str): The store_identifier which identifies this KeyValueStore instance. You can have multiple, separate KeyValueStore instances in a workspace using this implementation.
        """
        self.client = client
        self.store_identifier = f"kv-store-{store_identifier}"

    def _get_file(self, or_create: bool = False) -> Optional[File]:
        status_files = File.query(self.client, f'filetag and kind "{self.store_identifier}"').files
        if len(status_files) == 0:
            if not or_create:
                return None
            return File.create(
                self.client,
                blocks=[Block.CreateRequest(text="")],
                tags=[Tag.CreateRequest(kind=self.store_identifier, name="__init__")],
            )
        else:
            return status_files[0]

    def get(self, key: str) -> Optional[Dict]:
        """Gets the value represented by `key`."""
        file = self._get_file()

        if file is None:
            return None

        for tag in file.tags:
            if tag.kind == self.store_identifier and tag.name == key:
                return tag.value

    def delete(self, key: str) -> bool:
        """Deletes the entry represented by `key`"""
        file = self._get_file()

        if file is None:
            return False

        deleted = False
        for tag in file.tags:
            if tag.kind == self.store_identifier and tag.name == key:
                tag.delete()
                deleted = True

        return deleted

    def set(self, key: str, value: Dict[str, Any]):
        # Sets the entry (key, value)

        # First delete it if it exists to avoid duplicate tags.
        self.delete(key)

        # Now get/create the file
        file = self._get_file(or_create=True)

        req = Tag.CreateRequest(file_id=file.id, kind=self.store_identifier, name=key, value=value)
        return self.client.post("tag/create", req, expect=Tag)

    def reset(self):
        """Deletes all keys"""
        file = self._get_file()
        if file is not None:
            file.delete()



"""
Provides an example of how a third-party model can be incorporated into Steamship as a Trainable Tagger.

In this example, we have three classes:

- MockClient       -- Simulates the API client for a service such as Google AutoML
- ThirdPartyModel  -- Demonstrates how to build a model that is simply wrapping usage of the MockClient
- ThirdPartyTrainableTaggerPlugin -- Plugin-wrapper around the ThirdPartyModel
"""

import json
import logging
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Type

from assets.plugins.taggers.plugin_trainable_tagger import (
    TRAINING_PARAMETERS,
    TestTrainableTaggerModel,
)

from steamship import Block, File, SteamshipError, Tag
from steamship.base import Task, TaskState
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.train_status_plugin_input import TrainStatusPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.request import PluginRequest
from steamship.plugin.tagger import TrainableTagger
from steamship.plugin.trainable_model import TrainableModel

# If this isn't present, Localstack won't show logs
logging.getLogger().setLevel(logging.INFO)


class MockClient:
    """This is a simulation of what a third-party client to an AutoML-style API might look like.

    We simulate the fact that it offers:
    * The ability to upload a dataset
    * The ability to initiate an async model training job
    * The ability to check on that training job
    * The ability to run the resulting trained model.
    """

    class ThirdPartyTrainingStatus(str, Enum):
        TRAINING = "training"
        TRAINED = "trained"

    LABELS = ["label__mock"]
    FAKE_TRAINING_FILE_ID = "0000-0000-0000-0000"
    FAKE_MODEL_ID = "0000-0000-0000-0000"

    def upload_file_training_file(self, data: Optional[bytes] = None) -> str:
        """Pretends to upload a file and returns an ID representing that file on the remote third-party system"""
        return self.FAKE_TRAINING_FILE_ID

    def train(self, training_file_id: str) -> str:
        """Pretends to initiate a training process on the third-party system. Returns an ID representing the model ID."""
        return self.FAKE_MODEL_ID

    def training_status(self, model_id: str) -> ThirdPartyTrainingStatus:
        return self.ThirdPartyTrainingStatus.TRAINED

    def infer(self, text: str, model_id) -> List[str]:
        return self.LABELS


class ThirdPartyModel(TrainableModel):
    """Example of a trainable model that wraps a third-party training/inference process.

    We separate this class from the `MockClient` above because, in a real setting, the `MockClient` would be a
    pip-installable package (e.g. `pip install google-automl`), where as this class is the Steamship-created
    wrapper of that pacakge.

    From this *Steamship* perspective, the goal of this `ThirdPartyModel` wrapper class is to load and store
    all the necessary state for training and communicating with the remote model. This might be as simple as
    a single `MODEL_ID` parameter returned by the remote service, used for inquiring about training status
    and/or invoking the trained model.
    """

    # We save the the remote model ID in this file.
    PARAM_FILE = "model_information.json"
    params: Optional[Dict] = None
    client: Optional[MockClient] = None

    def __init__(self):
        self.client = MockClient()

    def load_from_folder(self, checkpoint_path: Path):
        """[Required by TrainableModel] Load state from the provided path."""
        with open(checkpoint_path / self.PARAM_FILE, "r") as f:
            self.params = json.loads(f.read())

    def save_to_folder(self, checkpoint_path: Path):
        """[Required by TrainableModel] Save state to the provided path."""
        with open(checkpoint_path / self.PARAM_FILE, "w") as f:
            f.write(json.dumps(self.params))

    def train(self, input: PluginRequest[TrainPluginInput]) -> InvocableResponse[TrainPluginOutput]:
        """Trains using the MockClient."""
        if self.client is None:
            raise SteamshipError(message="MockClient was null.")
        reference_data = {"num_checkins": 0}
        return InvocableResponse(
            status=Task(state=TaskState.running, remote_status_input=reference_data)
        )

    def train_status(
        self, input: PluginRequest[TrainStatusPluginInput]
    ) -> InvocableResponse[TrainPluginOutput]:
        logging.info(f'Called train_status with {input.status.remote_status_input["num_checkins"]}')
        input.status.remote_status_input["num_checkins"] += 1
        complete = input.status.remote_status_input["num_checkins"] > 2

        if not complete:
            return InvocableResponse(
                status=Task(
                    state=TaskState.running, remote_status_input=input.status.remote_status_input
                )
            )

        # Initialize the parameter bundle we'll eventually save to disk.
        self.params = {}

        # Step 1. Simulate preparing data
        data = None

        # Step 2. Simulate uploading data
        data_file_id = self.client.upload_file_training_file(data)
        self.params["data_file_id"] = data_file_id

        # Step 3. Simulate training
        model_id = self.client.train(data_file_id)
        self.params["model_id"] = model_id

        return InvocableResponse(data=TrainPluginOutput())

    def run(self, request: PluginRequest[BlockAndTagPluginInput]) -> BlockAndTagPluginOutput:
        """Runs the mock client"""

        if "model_id" not in self.params:
            raise SteamshipError(
                message="No model_id was found in model parameter file. Has the model been trained?"
            )

        output = BlockAndTagPluginOutput(file=File.CreateRequest(blocks=[]))

        for in_block in request.data.file.blocks:
            tags = self.client.infer(in_block.text, self.params["model_id"])
            out_block = Block.CreateRequest(
                id=in_block.id, tags=[Tag.CreateRequest(kind=tag) for tag in tags]
            )
            output.file.blocks.append(out_block)

        return output


class ThirdPartyTrainableTaggerPlugin(TrainableTagger):
    """Plugin Wrapper for the `ThirdPartyModel`.

    This wrapper class does little other than translate the plugin lifecycle requests from Steamship
    into calls on the `ThirdPartyModel` object above.
    """

    class EmptyConfig(Config):
        pass

    def config_cls(self) -> Type[Config]:
        return self.EmptyConfig

    def model_cls(self) -> Type[ThirdPartyModel]:
        return ThirdPartyModel

    def run_with_model(
        self, request: PluginRequest[BlockAndTagPluginInput], model: TestTrainableTaggerModel
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        """Downloads the model file from the provided workspace"""
        logging.debug(f"run_with_model {request} {model}")
        return InvocableResponse(json=model.run(request))

    def get_training_parameters(
        self, request: PluginRequest[TrainingParameterPluginInput]
    ) -> InvocableResponse[TrainingParameterPluginOutput]:
        logging.debug(f"get_training_parameters {request}")
        return InvocableResponse(data=TRAINING_PARAMETERS)

    def train(
        self, request: PluginRequest[TrainPluginInput], model: ThirdPartyModel
    ) -> InvocableResponse[TrainPluginOutput]:
        """Since trainable can't be assumed to be asynchronous, the trainer is responsible for uploading its own model file."""
        logging.debug(f"train {request}")
        return model.train(request)

    def train_status(
        self, request: PluginRequest[TrainStatusPluginInput], model: ThirdPartyModel
    ) -> InvocableResponse[TrainPluginOutput]:
        """Since trainable can't be assumed to be asynchronous, the trainer is responsible for uploading its own model file."""

        # Call train status
        response = model.train_status(request)

        if response.data:
            # We're done training. We need to save the result
            archive_path_in_steamship = model.save_remote(
                client=self.client, plugin_instance_id=request.context.plugin_instance_id
            )
            # Set the model location on the plugin output.
            response.data.archive_path = archive_path_in_steamship

        return response


handler = create_handler(ThirdPartyTrainableTaggerPlugin)



from typing import Type

from steamship.invocable import Config, Invocable, InvocableResponse, create_handler, post


class HelloWorld(Invocable):
    # Package configuration is always defined by a Pydantic object which derives from
    # the base `Config` object.
    class HelloWorldConfig(Config):
        greeting: str

        # The casing of config variables is preserved exactly as the Steamship package author defines them.
        # Snake case is permitted
        snake_case_config: str
        # Camel case is permitted
        camelCaseConfig: str  # noqa: N815
        # Defaulted from the Engine, rather than here.
        defaultConfig: str  # noqa: N815

    # This is unnecessary, since the base `Invocable` declares it, but by re-declaring it here
    # we get better type-hints from our editor.
    config: HelloWorldConfig

    @post("greet")
    def greet(self, name: str = "Person") -> InvocableResponse:
        greeting = self.config.greeting
        return InvocableResponse(string=f"{greeting}, {name}")

    @post("snake")
    def snake(self) -> InvocableResponse:
        return InvocableResponse(string=self.config.snake_case_config)

    @post("camel")
    def camel(self) -> InvocableResponse:
        return InvocableResponse(string=self.config.camelCaseConfig)

    @post("defaulted")
    def defaulted(self) -> InvocableResponse:
        return InvocableResponse(string=self.config.defaultConfig)

    def config_cls(self) -> Type[Config]:
        return self.HelloWorldConfig


handler = create_handler(HelloWorld)



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Optional

import inflection
from pydantic import HttpUrl

from steamship.base.model import CamelModel
from steamship.utils.utils import format_uri

DEFAULT_WEB_BASE = "https://app.steamship.com/"
DEFAULT_APP_BASE = "https://steamship.run/"
DEFAULT_API_BASE = "https://api.steamship.com/api/v1/"

ENVIRONMENT_VARIABLES_TO_PROPERTY = {
    "STEAMSHIP_API_KEY": "api_key",
    "STEAMSHIP_API_BASE": "api_base",
    "STEAMSHIP_APP_BASE": "app_base",
    "STEAMSHIP_WEB_BASE": "web_base",
    "STEAMSHIP_WORKSPACE_ID": "workspace_id",
    "STEAMSHIP_WORKSPACE_HANDLE": "workspace_handle",
}
DEFAULT_CONFIG_FILE = Path.home() / ".steamship.json"

# This stops us from including the `client` object in the dict() output, which is fine in a dict()
# but explodes if that dict() is turned into JSON. Sadly the `exclude` option in Pydantic doesn't
# cascade down nested objects, so we have to use this structure to catch all the possible combinations
EXCLUDE_FROM_DICT = {
    "client": True,
    "blocks": {"__all__": {"client": True, "tags": {"__all__": {"client": True}}}},
    "tags": {"__all__": {"client": True}},
}


class Configuration(CamelModel):
    api_key: str
    api_base: HttpUrl = DEFAULT_API_BASE
    app_base: HttpUrl = DEFAULT_APP_BASE
    web_base: HttpUrl = DEFAULT_WEB_BASE
    workspace_id: str = None
    workspace_handle: str = None
    profile: Optional[str] = None

    def __init__(
        self,
        config_file: Optional[Path] = None,
        **kwargs,
    ):
        # First set the profile
        kwargs["profile"] = profile = kwargs.get("profile") or os.getenv("STEAMSHIP_PROFILE")

        # Then load configuration from a file if provided
        config_dict = self._load_from_file(
            config_file or DEFAULT_CONFIG_FILE,
            profile,
            raise_on_exception=config_file is not None,
        )
        config_dict.update(self._get_config_dict_from_environment())
        kwargs.update({k: v for k, v in config_dict.items() if kwargs.get(k) is None})

        kwargs["api_base"] = format_uri(kwargs.get("api_base"))
        kwargs["app_base"] = format_uri(kwargs.get("app_base"))
        kwargs["web_base"] = format_uri(kwargs.get("web_base"))

        super().__init__(**kwargs)

    @staticmethod
    def _load_from_file(
        file: Path, profile: str = None, raise_on_exception: bool = False
    ) -> Optional[dict]:
        try:
            with file.open() as f:
                config_file = json.load(f)
                if profile:
                    if "profiles" not in config_file or profile not in config_file["profiles"]:
                        raise RuntimeError(f"Profile {profile} requested but not found in {file}")
                    config = config_file["profiles"][profile]
                else:
                    config = config_file
                return {inflection.underscore(k): v for k, v in config.items()}
        except FileNotFoundError:
            if raise_on_exception:
                raise Exception(f"Tried to load configuration file at {file} but it did not exist.")
        except Exception as err:
            if raise_on_exception:
                raise err
        return {}

    @staticmethod
    def _get_config_dict_from_environment():
        """Overrides configuration with environment variables."""
        return {
            property_name: os.getenv(environment_variable_name, None)
            for environment_variable_name, property_name in ENVIRONMENT_VARIABLES_TO_PROPERTY.items()
            if environment_variable_name in os.environ
        }



from __future__ import annotations

from typing import Any, Dict, Optional, Type, Union

from pydantic import BaseModel, Field

from steamship.base import Task
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.operations.tagger import TagRequest, TagResponse
from steamship.data.plugin import (
    HostingCpu,
    HostingEnvironment,
    HostingMemory,
    HostingTimeout,
    HostingType,
)
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput


class CreatePluginInstanceRequest(Request):
    id: str = None
    plugin_id: str = None
    plugin_handle: str = None
    plugin_version_id: str = None
    plugin_version_handle: str = None
    handle: str = None
    fetch_if_exists: bool = None
    config: Dict[str, Any] = None


class PluginInstance(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    plugin_id: str = None
    plugin_version_id: str = None
    workspace_id: Optional[str] = None
    user_id: str = None
    config: Dict[str, Any] = None
    hosting_type: Optional[HostingType] = None
    hosting_cpu: Optional[HostingCpu] = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_environment: Optional[HostingEnvironment] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["pluginInstance"] if "pluginInstance" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        plugin_id: str = None,
        plugin_handle: str = None,
        plugin_version_id: str = None,
        plugin_version_handle: str = None,
        handle: str = None,
        fetch_if_exists: bool = True,
        config: Dict[str, Any] = None,
    ) -> PluginInstance:
        """Create a plugin instance

        When handle is empty the engine will automatically assign one
        fetch_if_exists controls whether we want to re-use an existing plugin instance or not."""
        req = CreatePluginInstanceRequest(
            handle=handle,
            plugin_id=plugin_id,
            plugin_handle=plugin_handle,
            plugin_version_id=plugin_version_id,
            plugin_version_handle=plugin_version_handle,
            fetch_if_exists=fetch_if_exists,
            config=config,
        )

        return client.post("plugin/instance/create", payload=req, expect=PluginInstance)

    @staticmethod
    def get(client: Client, handle: str) -> PluginInstance:
        return client.post(
            "plugin/instance/get", IdentifierRequest(handle=handle), expect=PluginInstance
        )

    def tag(
        self,
        doc: Union[str, File],
    ) -> Task[
        TagResponse
    ]:  # TODO (enias): Should we remove this helper function in favor of always working with files?
        req = TagRequest(
            type="inline",
            file=File.CreateRequest(blocks=[Block.CreateRequest(text=doc)])
            if isinstance(doc, str)
            else doc,
            plugin_instance=self.handle,
        )
        return self.client.post(
            "plugin/instance/tag",
            req,
            expect=TagResponse,
        )

    def delete(self) -> PluginInstance:
        req = DeleteRequest(id=self.id)
        return self.client.post("plugin/instance/delete", payload=req, expect=PluginInstance)

    def train(self, training_request: TrainingParameterPluginInput) -> Task[TrainPluginOutput]:
        return self.client.post(
            "plugin/instance/train",
            payload=training_request,
            expect=TrainPluginOutput,
        )

    def get_training_parameters(
        self, training_request: TrainingParameterPluginInput
    ) -> TrainingParameterPluginOutput:
        return self.client.post(
            "plugin/instance/getTrainingParameters",
            payload=training_request,
            expect=TrainingParameterPluginOutput,
        )



from __future__ import annotations

from typing import Generic, Optional, TypeVar

from pydantic.generics import GenericModel

# Note!
# =====
#
# This the files in this package are for Plugin Implementors.
# If you are using the Steamship Client, you probably are looking for either steamship.client or steamship.data
#
from steamship.base import Task
from steamship.base.model import CamelModel, to_camel

T = TypeVar("T")
U = TypeVar("U")


class PluginRequestContext(CamelModel):
    """Contains the context in which"""

    plugin_id: str = None
    plugin_handle: str = None
    plugin_version_id: str = None
    plugin_version_handle: str = None
    plugin_instance_id: str = None
    plugin_instance_handle: str = None


class PluginRequest(GenericModel, Generic[T]):
    # The primary payload of the request. E.g. RawDataPluginInput, BlockAndTagPluginInput
    data: Optional[T] = None

    # The context in which this request is occurring
    context: Optional[PluginRequestContext] = None

    # The status of the request as perceived by the requester.
    status: Optional[Task] = None

    # Whether this plugin request is a status check against ongoing work. If True, status must be not None
    is_status_check: bool = False

    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True
